from pureml.utils.config import load_config
from collections import OrderedDict
import json
import time


def create_nodes(components):
    # print(components)

    nodes = [
        {
            "id": component["name"],
            "text": component["name"],
            "nodeType": component["type"],
            "desc": str(component["desc"]),
            "time": str(component["time"]),
        }
        for component in components
    ]

    return nodes


def create_extra_nodes(nodes, edges):
    # print(components)

    node_nodes = [i["id"] for i in nodes]
    edge_nodes = sum([[i["from"]] + [i["to"]] for i in edges], [])

    extra_nodes = list(set([n for n in edge_nodes if n not in node_nodes]))

    nodes = nodes + [
        {
            "id": n,
            "text": n,
            "nodeType": "unknown",
            "desc": str(None),
            "time": str(time.time()),
        }
        for n in extra_nodes
    ]

    return nodes


def create_edges(components):

    edges = []

    def add_node(node_from, node_to):
        edge_id = "--->".join([node_from, node_to])
        edge = {"id": edge_id, "from": node_from, "to": node_to}

        edges.append(edge)

    for n in components:
        if "parent" in n.keys():
            node_parent = n["parent"]
            node_to = n["name"]

            if node_parent is not None:

                if type(node_parent) == list:

                    for node_from in node_parent:
                        add_node(node_from, node_to)

                else:
                    add_node(node_parent, node_to)

    return edges


def prune_dict_elements(components):
    components_string = [json.dumps(d) for d in components]

    components_string = set(components_string)

    components_loaded = [json.loads(d) for d in components_string]

    return components_loaded


def create_lineage():
    config = load_config()

    load_data = config["load_data"]
    transformer = list(config["transformer"].values())
    dataset = config["dataset"]

    lineage_components = []

    if len(load_data) > 0:
        lineage_components.append(load_data)

    if len(transformer) > 0:
        lineage_components = lineage_components + transformer

    if len(dataset) > 0:
        lineage_components.append(dataset)

    lineage_components = prune_dict_elements(lineage_components)

    edges = create_edges(components=lineage_components)
    edges = prune_dict_elements(edges)

    nodes = create_nodes(components=lineage_components)
    nodes = create_extra_nodes(nodes, edges)
    nodes = prune_dict_elements(nodes)

    lineage = {"edges": edges, "nodes": nodes}

    return lineage


# const nodes = [
#   {
#     id: '1',
#     text: '1'
#   },
#   {
#     id: '2',
#     text: '2'
#   }
# ];

# const edges = [
#   {
#     id: '1-2',
#     from: '1',
#     to: '2'
#   }
# ];



from pydantic import BaseModel


class Singleton_BaseModel(BaseModel):
    _instance = None

    def __init__(self, **data):
        super().__init__(**data)
        if not self.__class__._instance:
            self.__class__._instance = self

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance



from pydantic import BaseModel, root_validator, validator
from .singleton import Singleton_BaseModel
import typing
import os


class PathSchema(Singleton_BaseModel):

    PATH_PUREML_RELATIVE = ".pureml"
    PATH_PREDICT_DIR_RELATIVE = "predict"

    PATH_USER_TOKEN = os.path.join(
        os.path.expanduser("~"), PATH_PUREML_RELATIVE, "token"
    )

    PATH_USER_PROJECT_DIR = os.path.join(os.getcwd(), PATH_PUREML_RELATIVE)

    PATH_USER_PROJECT = os.path.join(PATH_USER_PROJECT_DIR, "pure.project")

    PATH_CONFIG = os.path.join(PATH_USER_PROJECT_DIR, "config.pkl")  # 'temp.yaml'

    PATH_ARTIFACT_DIR = os.path.join(PATH_USER_PROJECT_DIR, "artifacts")
    PATH_ARRAY_DIR = os.path.join(PATH_USER_PROJECT_DIR, "array")
    PATH_AUDIO_DIR = os.path.join(PATH_USER_PROJECT_DIR, "audio")
    PATH_FIGURE_DIR = os.path.join(PATH_USER_PROJECT_DIR, "figure")
    PATH_TABULAR_DIR = os.path.join(PATH_USER_PROJECT_DIR, "tabular")
    PATH_VIDEO_DIR = os.path.join(PATH_USER_PROJECT_DIR, "video")
    PATH_IMAGE_DIR = os.path.join(PATH_USER_PROJECT_DIR, "image")

    PATH_DATASET_DIR = os.path.join(PATH_USER_PROJECT_DIR, "dataset")

    PATH_MODEL_DIR = os.path.join(PATH_USER_PROJECT_DIR, "model")

    PATH_PREDICT_DIR = os.path.join(PATH_USER_PROJECT_DIR, PATH_PREDICT_DIR_RELATIVE)

    class Config:
        arbitrary_types_allowed = True

    @root_validator(pre=False)
    def create_base_folders(cls, values):
        os.makedirs(values["PATH_USER_PROJECT_DIR"], exist_ok=True)

        return values

    @root_validator(pre=False)
    def create_log_folders(cls, values):
        os.makedirs(values["PATH_ARTIFACT_DIR"], exist_ok=True)
        os.makedirs(values["PATH_ARRAY_DIR"], exist_ok=True)
        os.makedirs(values["PATH_AUDIO_DIR"], exist_ok=True)
        os.makedirs(values["PATH_FIGURE_DIR"], exist_ok=True)
        os.makedirs(values["PATH_TABULAR_DIR"], exist_ok=True)
        os.makedirs(values["PATH_VIDEO_DIR"], exist_ok=True)
        os.makedirs(values["PATH_IMAGE_DIR"], exist_ok=True)

        return values

    @root_validator(pre=False)
    def create_model_folders(cls, values):
        os.makedirs(values["PATH_MODEL_DIR"], exist_ok=True)

        return values

    @root_validator(pre=False)
    def create_dataset_folders(cls, values):
        os.makedirs(values["PATH_DATASET_DIR"], exist_ok=True)

        return values

    @root_validator(pre=False)
    def create_predict_folders(cls, values):
        os.makedirs(values["PATH_PREDICT_DIR"], exist_ok=True)

        return values



import requests
from rich import print

import os
import json

from pureml.cli.helpers import get_auth_headers

from . import get_org_id
from pureml.schema import (
    ModelSchema,
    StorageSchema,
    ConfigKeys,
    ContentTypeHeader,
    AcceptHeader,
)
from pureml import save_model, load_model
from urllib.parse import urljoin
import joblib
from pureml.utils.hash import generate_hash_for_file
from pureml.utils.readme import load_readme
from pureml.utils.version_utils import parse_version_label
from pureml.utils.config import reset_config


config_keys = ConfigKeys
storage = StorageSchema().get_instance()


def init_branch(label):
    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/{}/branch/create".format(org_id, name)
    url = urljoin(model_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_JSON)

    data = {"model_name": name, "branch_name": branch}

    data = json.dumps(data)

    response = requests.post(url, data=data, headers=headers)

    if response.ok:
        print(f"[bold green]Branch has been created!")

        return True

    else:
        print(f"[bold red]Branch has not been created!")

        return False


def check_model_hash(hash: str, label: str):

    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/{}/branch/{}/hash-status".format(org_id, name, branch)
    url = urljoin(model_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    data = {"hash": hash, "branch": branch}

    data = json.dumps(data)

    response = requests.post(url, data=data, headers=headers)

    hash_exists = False

    if response.ok:
        hash_exists = response.json()["data"][0]

    return hash_exists


def branch_details(label: str):
    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/{}/branch/{}".format(org_id, name, branch)
    url = urljoin(model_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)
    # print(response.url)
    # print(response.text)

    if response.ok:
        # T-1161 standardize api response to contains Models as a list
        response_text = response.json()
        details = response_text["data"]
        # print(model_details)

        return details

    else:
        print(f"[bold red]Branch details details have not been found")
        return


def branch_status(label: str):

    details = branch_details(label=label)

    if details:
        return True
    else:
        return False


def branch_delete(label: str) -> str:
    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/{}/branch/{}/delete".format(org_id, name, branch)
    url = urljoin(model_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.delete(url, headers=headers)

    if response.ok:
        print(f"[bold green]Model branch has been deleted")

    else:
        print(f"[bold red]Unable to delete Model branch")

    return response.text


def branch_list(label: str) -> str:

    name, _, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/{}/branch".format(org_id, name)
    url = urljoin(model_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        response_text = response.json()
        branch_list = response_text["data"]

        return branch_list

    else:
        print(f"[bold red]Unable to obtain the list of branches!")

    return response.text


def list():
    """This function will return a list of all the modelst

    Returns
    -------
        A list of all the models

    """

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/all".format(org_id)
    url = urljoin(model_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print(f"[bold green]Obtained list of models")

        response_text = response.json()
        model_list = response_text["data"]
        # print(model_list)

        return model_list
    else:
        print(f"[bold red]Unable to obtain the list of models!")

    return


def init(label: str, readme: str = None):

    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    if readme is None:
        readme = ModelSchema().PATH_MODEL_README

    file_content, file_type = load_readme(path=readme)

    branch_user = "dev" if branch is None else branch
    branch_main = "main"

    url = "org/{}/model/{}/create".format(org_id, name)
    url = urljoin(model_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_JSON)

    data = {
        "name": name,
        "branch_names": [branch_main, branch_user],
        "readme": {"file_type": file_type, "content": file_content},
    }

    data = json.dumps(data)

    files = {"file": (readme, open(readme, "rb"), file_type)}

    response = requests.post(url, data=data, headers=headers)
    # response = requests.post(url, data=data, headers=headers, files=files)

    if response.ok:
        print(f"[bold green]Model has been created!")

        return True
    else:
        print(f"[bold red]Model has not been created!")

        return False


def register(
    model,
    label,
    is_empty: bool = False,
):

    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    model_file_name = ".".join([name, "pkl"])
    model_path = os.path.join(model_schema.paths.PATH_MODEL_DIR, model_file_name)

    os.makedirs(model_schema.paths.PATH_MODEL_DIR, exist_ok=True)

    save_model(model, name, model_path=model_path)

    model_hash = generate_hash_for_file(
        file_path=model_path, name=name, branch=branch, is_empty=is_empty
    )

    model_exists = model_status(label)

    if not model_exists:
        model_created = init(label)
        print("model_created", model_created)
        if not model_created:
            print("[bold red] Unable to register the model")
            return False, model_hash, "latest"

    branch_exists = branch_status(label)
    print("branch_exists", branch_exists)

    if not branch_exists:
        branch_created = init_branch(label)
        print("branch_created", branch_created)

        if not branch_created:
            print("[bold red] Unable to register the model")
            return False, model_hash, "latest"

    model_exists_remote = check_model_hash(hash=model_hash, label=label)

    if model_exists_remote:
        print(f"[bold red]Model already exists. Not registering a new version!")
        return True, model_hash, "latest"
    else:
        url = "org/{}/model/{}/branch/{}/register".format(org_id, name, branch)
        url = urljoin(model_schema.backend.BASE_URL, url)

        headers = get_auth_headers(content_type=None, accept=AcceptHeader.APP_JSON)

        files = {"file": (model_file_name, open(model_path, "rb"))}

        data = {
            "name": name,
            "branch": branch,
            "hash": model_hash,
            "is_empty": is_empty,
            "storage": storage.STORAGE,
        }

        response = requests.post(url, files=files, data=data, headers=headers)
        # print(response.request.headers)

        if response.ok:

            model_version = None

            # print(response.json())
            try:
                print(f"[bold green]Model has been registered!")

                model_version = response.json()["data"][0]["version"]
                print("Model Version: ", model_version)
                model_label = ":".join([label, model_version])
                print("Model label: ", model_label)

                # reset_config(key=config_keys.model.value)
            except Exception as e:
                print(
                    "[bold red] Incorrect json response. Model has not been registered"
                )
                print(e)

            return True, model_hash, model_version

        else:
            print(f"[bold red]Model has not been registered!")
            print(response.text)

        return False, model_hash, None


def model_status(label: str):

    name, _, _ = parse_version_label(label)

    model_details = details(label=label)

    if model_details:
        return True
    else:
        return False


def details(label: str):
    """It fetches the details of a model.

    Parameters
    ----------
    name : str
        The name of the model
    version: str
        The version of the model
    Returns
    -------
        The details of the model.

    """

    name, _, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/{}".format(org_id, name)
    url = urljoin(model_schema.backend.BASE_URL, url)

    headers = get_auth_headers(
        content_type=ContentTypeHeader.APP_FORM_URL_ENCODED,
        accept=AcceptHeader.APP_JSON,
    )

    response = requests.get(url, headers=headers)
    # print(response.url)
    # print(response.text)

    if response.ok:
        # print(f"[bold green]Model details have been fetched")
        response_text = response.json()
        model_details = response_text["data"][0]
        # print(model_details)

        return model_details

    else:
        print(f"[bold red]Model details have not been found")
        return


def version_details(label: str):
    """It fetches the details of a model.

    Parameters
    ----------
    name : str
        The name of the model
    version: str
        The version of the model
    Returns
    -------
        The details of the model.

    """

    name, branch, version = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/{}/branch/{}/version/{}".format(org_id, name, branch, version)
    url = urljoin(model_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print(f"[bold green]Model Version details have been fetched")
        response_text = response.json()
        model_details = response_text["data"][0]
        # print(model_details)

        return model_details

    else:
        print(f"[bold red]Model details have not been found")
        return


def fetch(label: str):
    """This function fetches a model from the server and returns it as a `Model` object

    Parameters
    ----------
    name : str, optional
        The name of the model you want to fetch.
    version: str
        The version of the model

    Returns
    -------
        The model is being returned.

    """

    name, branch, version = parse_version_label(label)

    org_id = get_org_id()

    model_details = version_details(label=label)

    if model_details is None:
        print(f"[bold red]Unable to fetch Model version")
        return

    is_empty = model_details["is_empty"]

    if is_empty:
        print("[bold orange]Model file is not registered to the version")
        return

    # storage_path = model_details["path"]["source_path"]
    # storage_source_type = model_details["path"]["source_type"]["public_url"]

    # model_url = urljoin(storage_source_type, storage_path)

    model_url = model_details["path"]

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(model_url)

    if response.ok:
        model_bytes = response.content
        open("temp_model.pure", "wb").write(model_bytes)

        model = load_model(model_path="temp_model.pure")

        # print(f"[bold green]Model version has been fetched")
        return model
    else:
        print(f"[bold red]Unable to fetch Model version")
        # print(response.status_code)
        # print(response.text)
        # print(response.url)
        return


# def delete(label: str) -> str:
#     """This function deletes a model from the project

#     Parameters
#     ----------
#     name : str
#         The name of the model you want to delete
#     version : str
#         The version of the model to delete.

#     """

#     name, _, _ = parse_version_label(label)

#     user_token = get_token()
#     org_id = get_org_id()
#     model_schema = ModelSchema()

#     url = "org/{}/model/{}/delete".format(org_id, name)
#     url = urljoin(model_schema.backend.BASE_URL, url)

#     headers = {
#         "Content-Type": ContentTypeHeader.APP_FORM_URL_ENCODED.value,
#         "Authorization": "Bearer {}".format(user_token),
#     }

#     response = requests.delete(url, headers=headers)

#     if response.ok:
#         print(f"[bold green]Model has been deleted")

#     else:
#         print(f"[bold red]Unable to delete Model")

#     return response.text


def serve_model():
    pass



from fastapi import Request, UploadFile
from pureml.predictor.predictor import BasePredictor
from pureml.schema.predict import Input, Output
import json
from .package import parse_input, parse_output
import shutil


async def predict_request_with_json(request: Request, predictor: BasePredictor):

    # input_type, input_shape = process_input(input=predictor.input)
    # output_type, output_shape = process_output(output=predictor.output)
    print("predictor.input", predictor.input)
    print("input_type", predictor.input.type)
    print("input_shape", predictor.input.shape)
    print("output_type", predictor.output.type)
    print("output_shape", predictor.output.shape)

    headers = request.headers
    body = await request.json()
    data_json = body["test_data"]
    print("data_json", data_json)

    data = parse_input(
        data=data_json,
        input_type=predictor.input.type,
        input_shape=predictor.input.shape,
    )

    if data is None:
        print("Error in data input format")
        predictions = json.dumps(None)
        return predictions

    predictions = predictor.predict(data)
    print("predictions", predictions)

    predictions = parse_output(
        data=predictions,
        output_type=predictor.output.type,
        output_shape=predictor.output.shape,
    )

    return predictions


async def predict_request_with_file(file: UploadFile, predictor: BasePredictor):

    # input_type, input_shape = process_input(input=predictor.input)
    # output_type, output_shape = process_output(output=predictor.output)
    print("predictor.input", predictor.input)
    print("input_type", predictor.input.type)
    print("input_shape", predictor.input.shape)
    print("output_type", predictor.output.type)
    print("output_shape", predictor.output.shape)

    if predictor.input.type == "None":
        print("Rebuild the docker container with non null input_type")
        predictions = json.dumps(None)
        return predictions

    path = None

    if not file:
        print("No upload file sent")
        predictions = json.dumps(None)
        return predictions
    else:
        path = file.filename
        print(path)
        with open(path, "wb") as buffer:
            shutil.copyfileobj(file.file, buffer)

    if path is None:
        print("Input image path is None")
        predictions = json.dumps(None)
        return predictions

    data = parse_input(
        data=path, input_type=predictor.input.type, input_shape=predictor.input.shape
    )

    if data is None:
        print("Error in data input format")
        predictions = json.dumps(None)
        return predictions

    predictions = predictor.predict(data)

    predictions = parse_output(
        data=predictions,
        output_type=predictor.output.type,
        output_shape=predictor.output.shape,
    )

    return predictions




from sklearn.metrics import recall_score


class Recall():

    def __init__(self):
        self.name = 'recall'
        self.input_type = 'int'
        self.output_type = None
        self.kwargs = None
        

    def parse_data(self, data):
        
        return data



    def compute(self, predictions, references, labels=None, pos_label=1, average="binary", 
                sample_weight=None, zero_division="warn", **kwargs):
        
        score = recall_score(y_true=references, y_pred=predictions, labels=labels, pos_label=pos_label, 
                             average=average, sample_weight=sample_weight, zero_division=zero_division)
        
        score = {
            self.name : float(score) if score.size == 1 else score
            }

        return score


from pydantic import BaseModel, root_validator

from pureml.schema.singleton import Singleton_BaseModel


class StorageSchema(Singleton_BaseModel):

    STORAGE: str = "PUREML-STORAGE"

    class Config:
        arbitrary_types_allowed = True



from pydantic import BaseModel
from pureml.predictor.predictor import BasePredictor
from .grade import Grader
from pureml.components import dataset
from typing import Any
from importlib import import_module
from rich import print


class Evaluator(BaseModel):
    task_type: str
    label_model: str
    label_dataset: str
    predictor: BasePredictor = None
    predictor_path: str = "predict.py"
    grader: Grader = None
    dataset: Any = None

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    def load_dataset(self):
        self.dataset = dataset.fetch(self.label_dataset)
        print("[bold green] Succesfully fetched the dataset")

    def load_predictor(self):
        module_path = self.predictor_path.replace(".py", "")
        module_import = import_module(module_path)

        predictor_class = getattr(module_import, "Predictor")

        self.predictor = predictor_class()
        print("[bold green] Succesfully fetched the predictor")

    def load_model(self):
        self.predictor.load_models()
        print("[bold green] Succesfully fetched the model")

    def load(self):
        self.load_dataset()
        self.load_predictor()
        self.load_model()

    def evaluate(self):
        pred = self.predictor.predict(self.dataset["x_test"])
        self.grader = Grader(task_type=self.task_type)
        values = self.grader.compute(
            references=self.dataset["y_test"], predictions=pred
        )

        return values


def eval(label_model: str, label_dataset: str, task_type: str):
    evaluator = Evaluator(
        task_type=task_type, label_dataset=label_dataset, label_model=label_model
    )

    evaluator.load()

    values = evaluator.evaluate()

    return values



# from pureml.utils.pipeline import add_predict_to_config 
# from pureml.utils.constants import PATH_PREDICT_REQUIREMENTS
# import shutil

# def predict(model_name:str, model_version:str, requirements_file:str=None):

#     def decorator(func):
        
#         def wrapper(*args, **kwargs):
#             print('Inside wrapper')
#             # 
#             func_output = func(*args, **kwargs)

#             return func_output


#         try:
#             add_predict_to_config(func=func, model_name=model_name, model_version=model_version, requirements_file=requirements_file)
#         except Exception as e:
#             print('Unable to add requirements to config')
#             print(e)


#         # print("Outside  wrapper")

#         return wrapper
#     # print('Outside decorator')
#         # 
#     return decorator


from pydantic import BaseModel, Field
import typing
from typing import List, Any
from abc import ABC, abstractmethod
from pureml.schema import Input, Output
from enum import Enum


class BasePredictor(BaseModel, ABC):
    label: str
    model: Any = None
    input: Input
    output: Output = Output()
    requirements_py: list = None
    requirements_sys: list = None

    class Config:
        arbitrary_types_allowed = True

    @abstractmethod
    def load_models(self):
        pass

    @abstractmethod
    def predict(self, **kwargs: typing.Any):
        pass

    # @abstractmethod
    def load_requirements_py(self):
        pass

    # @abstractmethod
    def load_requirements_sys(self):
        pass

    # @abstractmethod
    def load_resources(self):
        pass



import os

def create_readme():
    file_content = 'Readme.md'
    file_type = 'md'

    return file_content, file_type

def load_readme(path:str=None):

    file_content, file_type = create_readme()


    if path is None:
        return file_content, file_type


    if os.path.exists(path):
        with open(path, 'r') as f:
            try:
                file_content = f.read()
                file_type = path.rsplit('.',1)[-1]
            except Exception as e:
                print('Unable to read the ReadME file.')
                print('Creating an Empty ReadME file')
    else:
        with open(path, 'w') as f:
            f.write(file_content)
        print('ReadME file does not exist.')
        print('Creating an Empty ReadME file')
        

        
    return file_content, file_type
        


import typer
from pureml.cli.puremlconfig import PureMLConfigYML
from pureml.config.parser import Config
# from pureml.trainer.train import Trainer
import os

app = typer.Typer()


import typer
from rich import print
# from puretrainer.train import Trainer
import os
from dotenv import load_dotenv

load_dotenv()


import pureml.cli.auth as auth
import pureml.cli.secrets as secrets
import pureml.cli.orgs as orgs

app = typer.Typer()
app.add_typer(auth.app, name="auth")
app.add_typer(secrets.app, name="secrets")
app.add_typer(orgs.app, name="orgs")


@app.callback(no_args_is_help=True)
def validate_user_authentication(ctx: typer.Context):
    # print(ctx.invoked_subcommand)
    if ctx.invoked_subcommand in ['auth']:
        return
    # user_token = auth.auth_validate()
    return


# init the config file
@app.command()
def init(
    silent: bool = typer.Option(
        False, "--silent", "-s", help="Run in silent mode"
    ),
    # repository: str = typer.Option(
    #     None, "--repository", "-r", help="The repository to use"
    # ),
    backend_url: str = typer.Option(
        None, "--backend-url", "-b", help="The backend url to use"
    ),
    frontend_url: str = typer.Option(
        None, "--frontend-url", "-f", help="The frontend url to use"
    ),
):
    """
    This command will initialize the puremlconfig.yaml file for your project
    """
    from pathlib import Path

    project_path = Path.cwd()

    puremlconfig = PureMLConfigYML(project_path / "puremlconfig.yaml")
    
    # check if the config file exists
    if puremlconfig.file.exists():
        print("A puremlconfig.yaml file already exists in this directory - aborting")
        return
    
    # if not silent, ask for the config values
    if not silent:
        print("Let's get started with the configuration of your project")

        # if not repository:
        #     repository = typer.prompt("Enter your preferred storage repository path [eg. file://pureml_data/]")
        # else:
        #     print(f"Using given repository: {repository}")

        isSelfHosted = typer.confirm("Are you using a self-hosted PureML instance?")

        if isSelfHosted:
            if not backend_url:
                backend_url = typer.prompt("Enter the backend url", default="http://localhost:8080/api/")
            else:
                print(f"Using given backend url: {backend_url}")
            
            if not frontend_url:
                frontend_url = typer.prompt("Enter the frontend url", default="http://localhost:3000/")
            else:
                print(f"Using given frontend url: {frontend_url}")
    else:
        # if not repository:
        #     print("Missing required arguments")
        #     return
        # print(f"Using given repository: {repository}")
        if backend_url:
            print(f"Using given backend url: {backend_url}")
        if frontend_url:
            print(f"Using given frontend url: {frontend_url}")
    
    # create the config file
    data = {
        # "repository": repository,
    }
    if backend_url:
        data["backend_url"] = backend_url
    if frontend_url:
        data["frontend_url"] = frontend_url
    puremlconfig.save(data)


from pydantic import BaseModel

from .backend import BackendSchema
from .paths import PathSchema
import os


class DatasetSchema(BaseModel):
    paths: PathSchema = PathSchema().get_instance()
    backend: BackendSchema = BackendSchema().get_instance()

    PATH_DATASET_README = os.path.join(paths.PATH_DATASET_DIR, "ReadME.md")



from .classification import Classification
from .regression import Regression
from pydantic import BaseModel, root_validator
import typing
from typing import Literal
from enum import Enum


class TaskTypes(str, Enum):
    classification = "classification"
    regression = "regression"


class Grader(BaseModel):
    # def __init__(self, task_type):
    task_type: TaskTypes

    kwargs: typing.Any = None
    scores: dict = {}
    task_grader: typing.Any = None

    # print(self.task_grader)

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    @root_validator(pre=True)
    def _set_fields(cls, values: dict) -> dict:

        task_type = values["task_type"]

        if task_type is not None:

            if task_type == TaskTypes.classification:
                task_grader = Classification()
            elif task_type == TaskTypes.regression:
                task_grader = Regression()
            else:
                task_grader = None

            values["task_grader"] = task_grader

        return values

    def compute(self, references, predictions, prediction_scores=None, **kwargs):

        # print(self.task_grader)
        if self.task_grader is not None:
            # print('task grader is not none')
            self.task_grader.kwargs = kwargs
            self.task_grader.references = references
            self.task_grader.predictions = predictions
            self.task_grader.prediction_scores = prediction_scores

            self.scores = self.task_grader.compute()

        return self.scores


def grader(task_type, **kwargs):
    grade = Grader(task_type=task_type)

    return grade



from .accuracy import Accuracy
from .confusion_matrix import ConfusionMatrix
from .f1 import F1
from .mae import MAE
from .mse import MSE
from .precision import Precision
from .recall import Recall
from .roc_auc import ROC_AUC



import os
import docker
from .fastapi import create_fastapi_file
from pureml.schema import FastAPISchema, PredictSchema, DockerSchema
import string, random
from pureml.utils.env import generate_env_dict, validate_env_docker

prediction_schema = PredictSchema()
fastapi_schema = FastAPISchema()
docker_schema = DockerSchema()


def generate_sys_commands(sys_commands):
    commands = "\nRUN apt update\n"
    if len(sys_commands) > 0:
        sys_commands = ["RUN " + i for i in sys_commands]
        commands += "\n".join(sys_commands)

    return commands


def create_docker_file(sys_commands):
    # os.makedirs(PATH_DOCKER_DIR, exist_ok=True)
    os.makedirs(prediction_schema.paths.PATH_PREDICT_DIR, exist_ok=True)

    req_pos = prediction_schema.PATH_PREDICT_REQUIREMENTS.rfind(
        prediction_schema.REQUIREMENTS_NAME
    )
    req_path = prediction_schema.PATH_PREDICT_REQUIREMENTS[req_pos:]

    api_pos = fastapi_schema.PATH_FASTAPI_FILE.rfind(
        prediction_schema.paths.PATH_PREDICT_DIR_RELATIVE
    )
    api_path = fastapi_schema.PATH_FASTAPI_FILE[api_pos:]

    commands = generate_sys_commands(sys_commands=sys_commands)

    docker = """
    
FROM {BASE_IMAGE}

{commands}

RUN mkdir -p {PREDICT_DIR}

WORKDIR {PREDICT_DIR}

ADD . {PREDICT_DIR}

COPY . .

RUN pip install --no-cache-dir --upgrade pip \
  && pip install --no-cache-dir -r {REQUIREMENTS_PATH}

RUN pip install fastapi uvicorn python-multipart

RUN pip install pureml

EXPOSE {PORT}
CMD ["python", "{API_PATH}"]    
""".format(
        BASE_IMAGE=docker_schema.BASE_IMAGE_DOCKER,
        PORT=docker_schema.PORT_DOCKER,
        PREDICT_DIR=prediction_schema.paths.PATH_PREDICT_DIR_RELATIVE,
        API_PATH=api_path,
        REQUIREMENTS_PATH=req_path,
        commands=commands,
    )

    with open(docker_schema.PATH_DOCKER_IMAGE, "w") as docker_file:
        docker_file.write(docker)

    docker_yaml = """version: '3'

services:
  prediction:
    build: .
    container_name: "{CONTAINER_NAME}"
    expose:
      - "{DOCKER_PORT}"
    ports:
      - "{HOST_PORT}:{DOCKER_PORT}"
    
    """.format(
        DOCKER_PORT=docker_schema.PORT_DOCKER,
        HOST_PORT=docker_schema.PORT_HOST,
        CONTAINER_NAME="pureml_prediction",
    )

    with open(docker_schema.PATH_DOCKER_CONFIG, "w") as docker_yaml_file:
        docker_yaml_file.write(docker_yaml)


def create_docker_image(label):
    # if image_tag is None:
    # image_tag = "pureml_docker_image"
    label = label.replace(" ", "")

    image_repository = label.split(":", 1)[0]
    image_tag = label.split(":", 1)[1].replace(":", "-")

    # random_value = "".join(random.choices(string.ascii_lowercase + string.digits, k=8))
    # image_tag = "-".join([image_tag, random_value])

    image_tag = ":".join([image_repository, image_tag])

    client = docker.from_env()

    docker_file_path_relative = docker_schema.PATH_DOCKER_IMAGE.split(os.path.sep)[-1]
    # print(docker_schema.paths.PATH_PREDICT_DIR)
    # print(docker_file_path_relative)
    # print(image_tag)

    try:
        image, build_log = client.images.build(
            path=docker_schema.paths.PATH_PREDICT_DIR,
            dockerfile=docker_file_path_relative,
            tag=image_tag,
            nocache=True,
            rm=True,
        )

        print("Docker image is created.")
        print("Tag: ", image_tag)
        print("Image Long Id:", image.id)
        print("Image Short Id: ", image.short_id)

    except Exception as e:
        print(e)
        image = None
        build_log = None

    return image, build_log, image_tag


def run_docker_container(image, runtime, gpu_ids, host_port, docker_port, env_dict):
    client = docker.from_env()
    name = image.tags[0].replace(":", "-")

    random_value = "".join(random.choices(string.ascii_lowercase + string.digits, k=8))
    name = "-".join([name, random_value])

    docker_port = "{p}/tcp".format(p=docker_port)

    # print("docker_port", docker_port)
    container_args = {
        "image": image,
        "ports": {docker_port: host_port},
        "detach": True,
        "auto_remove": True,
        "name": name,
        "runtime": runtime,
        "environment": env_dict,
    }

    if len(gpu_ids) != 0:
        gpu_ids = [str(id) for id in gpu_ids]
        container_args["device_requests"] = [
            docker.types.DeviceRequest(device_ids=gpu_ids, capabilities=[["gpu"]])
        ]

    container = client.containers.run(**container_args)

    # container = client.containers.run(
    #     image=image,
    #     ports={docker_port: docker_schema.PORT_HOST},
    #     detach=True,
    #     name=name,
    #     runtime=runtime,
    # )

    return container


def create(
    label,
    # org_id,
    # access_token,
    env_path,
    runtime=None,
    gpu_ids=[],
    port=None,
    predict_path=None,
    requirements_path=None,
    sys_commands=[],
):

    env_dict = generate_env_dict(env_path=env_path)
    env_valid = validate_env_docker(env_dict=env_dict)

    if not env_valid:
        return None

    create_fastapi_file(
        label=label, predict_path=predict_path, requirements_path=requirements_path
    )

    create_docker_file(sys_commands=sys_commands)

    image, build_log, image_tag = create_docker_image(label=label)

    if image is not None:

        if port is None:
            host_port = docker_schema.PORT_HOST
        else:
            host_port = port

        docker_port = docker_schema.PORT_DOCKER

        container = run_docker_container(
            image=image,
            runtime=runtime,
            gpu_ids=gpu_ids,
            host_port=host_port,
            docker_port=docker_port,
            env_dict=env_dict,
        )

        print("Created Docker container")
        print("Container Name: ", container.name)
        print("Container Long Id: ", container.id)
        print("Container id: ", container.short_id)

        print(
            "Prediction requests can be forwarded to {ip}:{p}/predict".format(
                ip=docker_schema.API_IP_HOST, p=host_port
            )
        )
    else:
        print("Failed to create the container")
        print("Build Log")
        print(build_log)


def get(container_id):

    client = docker.from_env()

    container = client.containers.get(container_id)

    return container


def stop(container_id):

    client = docker.from_env()

    container = client.containers.get(container_id)

    container.stop()



from typing import Dict, List
import requests
import typer
from rich import print
from rich.console import Console
from rich.table import Table
from urllib.parse import urljoin
from pureml.cli.helpers import get_auth_headers, save_auth
from pureml.schema import BackendSchema, PathSchema, ContentTypeHeader, AcceptHeader
from pureml.schema.backend import get_backend_base_url

path_schema = PathSchema().get_instance()
backend_schema = BackendSchema().get_instance()
app = typer.Typer()


@app.callback()
def callback():
    """
    Organization command

    Use with show or select option

    show - lists all the organizations\n
    select - select an organization
    """


@app.command()
def show(
    backend_url: str = typer.Option(
        "",
        "--backend-url",
        "-b",
        help="Backend URL for self-hosted or custom pureml backend instance",
    )
):
    url_path = "org"
    url = urljoin(get_backend_base_url(backend_url), url_path)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_JSON)
    response = requests.get(url, headers=headers)
    if response.ok:
        print()
        print("[bold green] You are part of following Organizations!")
        org_all = response.json()["data"]
        console = Console()
        count = 0
        table = Table("Sr.No.", "Name", "Description")
        org_data = {}
        for org in org_all:
            count += 1
            table.add_row(str(count), org["org"]["name"], org["org"]["description"])
            org_data[count] = {"id": org["org"]["uuid"], "name": org["org"]["name"]}

        console.print(table)
        print()
        return count, org_data
    else:
        print("[bold red]Unable to get the list of organizations!")
        return None, None


@app.command()
def select(
    backend_url: str = typer.Option(
        "",
        "--backend-url",
        "-b",
        help="Backend URL for self-hosted or custom pureml backend instance",
    )
):
    print("[bold green] Select the Sr.No. of Organization")
    count, org_data = show(backend_url=backend_url)
    if count:
        sr_no = -1
        while int(sr_no) not in range(1, count + 1):
            sr_no: str = typer.prompt(
                "Enter your Sr.No. of Organization (1 .... " + str(count) + ")"
            )
            if int(sr_no) not in range(1, count + 1):
                print("[bold red]Invalid Sr.No. of Organization!")
        save_auth(org_id=org_data[int(sr_no)]["id"])
        selected_org_name = org_data[int(sr_no)]["name"]
        print(f"[green]Successfully linked to organization {selected_org_name}!")
    else:
        print("[bold red]Did not Select any organization!")
        return None


# Possibly useful for future commands
# Moved from auth.py
def check_org_status(headers: dict, base_url: str):
    """
    Headers: dict
        headers from get_auth_headers() invocation of parent function

    base_url: str
        base_url from get_backend_base_url() invocation of parent function
    """

    org_id: str = typer.prompt("Enter your Org Id")

    url_path = "org/id/{}".format(org_id)
    url = urljoin(base_url, url_path)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print("[green]Organization Exists!")
        return org_id
    else:
        print("[red]Organization does not Exists!")
        return None



from distutils.command.config import config
import yaml
from pydantic import BaseModel
import typing
import pandas as pd
import pyarrow.parquet as pq
import os
from collections import OrderedDict, defaultdict

class Config(BaseModel):
    config_path: str = './pureml.yaml'
    config: dict = None

    input_parameters: typing.Any = None
    output_parameters: typing.Any = None
    label_header : typing.Any = None

    data_path: typing.Any = None
    project_folder: typing.Any = None

    engine: str = None
    model: str = None
    model_parameters: typing.Any = None
    optimize: str = None

    integrations: dict = None



    def create_config(self, project, label, data):
        config_dict = OrderedDict()
        # config_dict = defaultdict()

        config_dict['Version'] = 0.1

        config_dict['Organization'] = {
                    'project' : {
                        'name' : project, 
                        'version' : 1
                    },
                    'data' : {
                        'name' : 'Project Data',
                        'type' : 'url',
                        'format' : 'parquet',
                        'path' : data
                    }
        }


        parquet_file = pq.read_table(data, columns=['A']).to_pandas()
        
        config_dict['Output paramters'] = {
                        'name' : label,
                        'type' : str(parquet_file.dtypes[label])
        }


        config_dict = dict(config_dict)


        with open(os.path.join(project, 'config.yaml'), "w") as f:
            yaml.dump(config_dict, f, sort_keys=False)



    def load_config(self):
        self.config = yaml.load(open(self.config_path, 'r'), Loader=yaml.SafeLoader)

        # self.data_path = self.config['Organization']['data']['path']
        self.project_folder = self.config['Organization']['project']['name']
        # os.makedirs(self.project_folder, exist_ok=True)


        # self.label_header = self.config['Output paramters']['name']

        # if 'trainer' in self.config.keys():
        #     if 'engine' in self.config['trainer'].keys():
        #         self.engine = self.config['trainer']['engine']

        #     if 'model' in self.config['trainer'].keys():
        #         if 'name' in self.config['trainer']['model'].keys():
        #             self.model = self.config['trainer']['model']['name']

        #         if 'parameters' in self.config['trainer']['model'].keys():
        #             self.model_parameters = self.config['trainer']['model']['parameters']


        # if 'optimizer' in self.config.keys():
        #     if 'name' in self.config['optimizer'].keys():
        #         name = self.config['optimizer']['name']
        #         if name == 'optuna':
        #             self.optimize = True


        self.load_integrations()        

    def load_integrations(self):
        if 'integrations' in self.config.keys():
            self.integrations = self.config['integrations']
            # if 'experiment' in integrations.keys():
        


    def generate_config(self):

        parquet_file = pq.read_table(self.data_path)
        file_metadata = parquet_file.schema.pandas_metadata


        columns_input = []

        for column in file_metadata['columns']:
            if column['name'] == column['field_name']:
                columns_input.append({'name': column['name'], 'type':column['numpy_type']})

        
        self.config['Input paramters'] = columns_input

        # self.config

        with open(os.path.join(self.project_folder, 'config.yaml'), "w") as f:
                # json.dump(project_params, f)
            # print(project_params)
            yaml.dump(self.config, f, sort_keys=False)




        






import os
import joblib
from collections import defaultdict, OrderedDict
from pureml.schema import PathSchema, ConfigKeys


path_schema = PathSchema().get_instance()
config_keys = ConfigKeys


def load_config():
    os.makedirs(path_schema.PATH_USER_PROJECT_DIR, exist_ok=True)

    if os.path.exists(path_schema.PATH_CONFIG):
        config = joblib.load(path_schema.PATH_CONFIG)
    else:
        config = defaultdict()

        config[config_keys.load_data.value] = defaultdict()
        config[config_keys.transformer.value] = OrderedDict()
        config[config_keys.dataset.value] = defaultdict()

        config[config_keys.model.value] = OrderedDict()

        config[config_keys.params.value] = defaultdict()
        config[config_keys.metrics.value] = defaultdict()
        config[config_keys.figure.value] = defaultdict()
        config[config_keys.artifacts.value] = defaultdict()

        config[config_keys.pred_function.value] = defaultdict()
        config[config_keys.pip_requirement.value] = defaultdict()
        config[config_keys.resource.value] = defaultdict()

        joblib.dump(config, path_schema.PATH_CONFIG)

    return config


def save_config(config):
    save_dir = os.path.dirname(path_schema.PATH_CONFIG)

    os.makedirs(save_dir, exist_ok=True)
    # print('config')
    # print(config)

    joblib.dump(config, path_schema.PATH_CONFIG)


def reset_config(key):
    if os.path.exists(path_schema.PATH_CONFIG):
        config = joblib.load(path_schema.PATH_CONFIG)
        if key in config.keys():
            if key in [config_keys.transformer.value, config_keys.model.value]:
                config[key] = OrderedDict()
            else:
                config[key] = defaultdict()
            save_config(config)



from pathlib import Path
from typing import Dict

from ..model_framework import ModelFramework, ModelFrameworkType
from ..packaging_utils import infer_requirements




class Pytorch(ModelFramework):
    framework_name: str = 'torch'
    additional_requirements: list = [
                                        'torchvision',
                                        'torchaudio',
                                        'numpy',
                                        'joblib'
                                    ]
    requirements: list = [framework_name]


    def typ(self) -> ModelFrameworkType:
        return ModelFrameworkType.PYTORCH


    def supports_model_class(self, model_class) -> bool:
        model_framework, _, _ = model_class.__module__.partition(".")
        return model_framework == ModelFrameworkType.PYTORCH.value

    def get_requirements(self):

        default_requirements = [infer_requirements(framework_name=self.framework_name)]

        self.requirements = default_requirements + self.additional_requirements

        return self.requirements


import os, json
from pureml.schema import PathSchema
from rich import print

path_schema = PathSchema().get_instance()


def get_token():
    """It checks if the token exists in the user's home directory. If it does, it returns the token. If it
    doesn't, it returns None

    Returns
    -------
        The token is being returned.
    """
    path = path_schema.PATH_USER_TOKEN
    # path = os.path.expanduser(path)

    if os.path.exists(path):
        creds = open(path, "r").read()

        creds_json = json.loads(creds)
        if "accessToken" in creds_json:
            token = creds_json["accessToken"]
            return token
    return


def get_api_token():
    """It checks if the api token exists in the user's home directory. If it does, it returns the api token id and key. If it
    doesn't, it returns None

    Returns
    -------
        The api token id and key is returned.
        Dictionary format: {"api_id": api_id, "api_key": api_key}
    """
    path = path_schema.PATH_USER_TOKEN
    # path = os.path.expanduser(path)

    if os.path.exists(path):
        creds = open(path, "r").read()

        creds_json = json.loads(creds)
        if "api_id" in creds_json and "api_key" in creds_json:
            api_id = creds_json["api_id"]
            api_key = creds_json["api_key"]
            return {
                "api_id": api_id,
                "api_key": api_key
            }
    return


def delete_token(silent=False):
    """It checks if the token exists in the user's home directory. If it does, it deletes the token. If it
    doesn't, it returns None

    Returns
    -------
        The token is being returned.
    """
    path = path_schema.PATH_USER_TOKEN
    # path = os.path.expanduser(path)

    if os.path.exists(path):
        os.remove(path)
        if not silent:
            print(f"[bold green]Authentication token deleted!")
        return
    else:
        if not silent:
            print(f"[bold red]Authentication token does not exist! Please login")
        return


def get_org_id():
    """It checks if the org exists in the user's home directory. If it does, it returns the org. If it
    doesn't, it returns None

    Returns
    -------
        The org is being returned.

    """

    path = path_schema.PATH_USER_TOKEN

    path = os.path.expanduser(path)

    if os.path.exists(path):
        creds = open(path, "r").read()

        creds_json = json.loads(creds)

        if "org_id" in creds_json:
            org_id = creds_json["org_id"]
            # print(f"[bold green]Organization exists!")

            # print(org_id)
            return org_id
        
    print(f"[bold red]Organization id does not exist! Please login")
    return


def convert_values_to_string(logged_dict):

    for key in logged_dict:
        logged_dict[key] = str(logged_dict[key])

    return logged_dict



from collections import defaultdict
import hashlib
import string
import random
from .config import load_config, save_config
import json
import inspect
import re
from datetime import datetime
from pureml.components import get_org_id


def file_reader_chunk(file_obj, chunk_size=1024):
    """Generator that reads a file in chunks of bytes"""
    while True:
        chunk = file_obj.read(chunk_size)
        if not chunk:
            return
        yield chunk


def generate_hash_unique(name, branch, hash=hashlib.md5):

    org_id = get_org_id()

    time_current = str(datetime.now())

    string_random = "".join(
        random.choices(string.ascii_lowercase + string.digits, k=16)
    )

    hash_content = "puremlHash".join(
        [org_id, name, branch, time_current, string_random]
    )

    # print('hash content', hash_content)

    value_str = hash_content.encode()
    # print('value_str', value_str)

    file_object = hash(value_str)
    # print(file_object)

    hash_value = file_object.hexdigest()

    return hash_value


def generate_hash_for_file(
    file_path, name: str, branch: str, hash=hashlib.md5, is_empty=False
):

    if is_empty:
        hash_value = generate_hash_unique(name=name, branch=branch)
    else:
        hash_obj = hash()
        file_object = open(file_path, "rb")

        for chunk in file_reader_chunk(file_object):
            hash_obj.update(chunk)

        hash_value = hash_obj.hexdigest()

        file_object.close()

    return hash_value


def generate_hash_for_dict(values, hash=hashlib.md5):
    value_str = json.dumps(values).encode()

    file_object = hash(value_str)

    hash_value = file_object.hexdigest()

    return hash_value


def generate_hash_for_function(func, hash=hashlib.md5):

    string_stripped = re.sub(r"[\n\t\s]*", "", inspect.getsource(func))

    string_stripped = string_stripped.encode()

    string_object = hash(string_stripped)

    hash_value = string_object.hexdigest()

    return hash_value


# def check_hash_status_model(hash_value, name, item_key='model'):

#     hash_status = False

#     config = load_config()
#     if len(config) > 0:
#         models = config[item_key]
#         for model_invoke_order, model_details in models.items():
#             # print('model_invoke_order', model_invoke_order)
#             # print('model_details', model_details)

#             if hash_value == model_details['hash']:
#                 hash_status = True
#                 return hash_status, hash_value


#     return hash_status


# def check_hash_status_dataset(file_path, name, item_key='dataset'):
#     hash_value = generate_hash_for_file(file_path=file_path)
#     hash_status = False

#     config = load_config()
#     if len(config) > 0:
#         dataset = config[item_key]


#         if hash_value == dataset['hash']:
#             hash_status = True
#             return hash_status, hash_value


#     return hash_status, hash_value



import json
from urllib.parse import urljoin

import requests
from pureml.cli.helpers import get_auth_headers
from pureml.schema import BackendSchema, LogSchema, ConfigKeys, ContentTypeHeader
from pureml.utils.log_utils import merge_step_with_value
from pureml.utils.pipeline import add_metrics_to_config
from rich import print

from . import convert_values_to_string, get_org_id
from pureml.utils.version_utils import parse_version_label
from pureml.utils.config import reset_config


backend_schema = BackendSchema().get_instance()
post_key_predict = LogSchema().key.metrics.value
config_keys = ConfigKeys


def post_metrics(metrics, model_name: str, model_branch: str, model_version: str):

    org_id = get_org_id()

    url = "org/{}/model/{}/branch/{}/version/{}/log".format(
        org_id, model_name, model_branch, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.ALL)

    metrics = json.dumps(metrics)
    data = {"data": metrics, "key": post_key_predict}

    data = json.dumps(data)

    response = requests.post(url, data=data, headers=headers)

    if response.ok:
        print(f"[bold green]Metrics have been registered!")
        reset_config(key=config_keys.metrics.value)

    else:
        print(f"[bold red]Metrics have not been registered!")

    return response


def add(
    metrics,
    label: str = None,
    step=1,
) -> str:
    """`add()` takes a dictionary of metrics and a model name as input and returns a string

    Parameters
    ----------
    metrics
        a dictionary of metrics
    model_name : str
        The name of the model you want to add metrics to.
    model_version: str
        The version of the model

    Returns
    -------
        The response.text is being returned.

    """
    model_name, model_branch, model_version = parse_version_label(label)

    metrics = convert_values_to_string(logged_dict=metrics)
    # metrics = merge_step_with_value(values_dict=metrics, step=step)

    add_metrics_to_config(
        values=metrics,
        model_name=model_name,
        model_branch=model_branch,
        model_version=model_version,
    )

    if (
        model_name is not None
        and model_branch is not None
        and model_version is not None
    ):
        response = post_metrics(
            metrics=metrics,
            model_name=model_name,
            model_branch=model_branch,
            model_version=model_version,
        )

        # return response.text

    # return


def details(label: str):
    model_name, model_branch, model_version = parse_version_label(label)
    org_id = get_org_id()

    url = "org/{}/model/{}/branch/{}/version/{}/log".format(
        org_id, model_name, model_branch, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # T-1161 standardize api response to contains Models as a list
        response_text = response.json()
        details = response_text["data"]

        # print(model_details)

        return details

    else:
        print(f"[bold red]Unable to fetch logs!")
        return


def get_value_from_log(details, key_log=post_key_predict, key=None):
    value = None
    if details is not None:
        # print(details)

        for det in details:
            # print(det)
            # print(det["key"])
            if det["key"] == key_log:
                value = det["data"]
                # print(value)
                value = json.loads(value)

                if key is not None:
                    if key in value.keys():
                        value = value[key]
                    else:
                        print(
                            "[bold red]{} : {} is not available for the model!".format(
                                key_log, key
                            )
                        )

                return value

    print("[bold red] Unable to find the {}".format(key))

    return value


def fetch(label: str, metric: str = None) -> str:
    """This function fetches the metrics of a model

    Parameters
    ----------
    model_name : str
        The name of the model you want to fetch metrics for.
    model_version: str
        The version of the model
    metric : str
        The metric you want to fetch. If you want to fetch all the metrics, leave this parameter empty.

    Returns
    -------
        The metrics that are fetched

    """

    metric_details = details(label=label)

    if metric_details:

        metrics = get_value_from_log(
            details=metric_details, key_log=post_key_predict, key=metric
        )

        return metrics

    return


# def delete(metric: str, label: str) -> str:
#     """This function deletes a metric from a model

#     Parameters
#     ----------
#     model_name : str
#         The name of the model you want to delete the metric from
#     metric : str
#         The name of the metric to delete
#     model_version: str
#         The version of the model

#     """
#     model_name, model_branch, model_version = parse_version_label(label)

#     user_token = get_token()
#     org_id = get_org_id()
#     log_schema = LogSchema()

#     url = "org/{}/model/{}/branch/{}/version/{}/log/delete".format(
#         org_id, model_name, model_branch, model_version
#     )
#     url = urljoin(log_schema.backend.BASE_URL, url)

#     headers = {
#         "Content-Type": "application/x-www-form-urlencoded",
#         "Authorization": "Bearer {}".format(user_token),
#     }

#     response = requests.delete(url, headers=headers)

#     if response.status_code == 200:
#         print(f"[bold green]Metric has been deleted")

#     else:
#         print(f"[bold red]Unable to delete Metric")

#     return response.text



import os

from .packaging import load_model, save_model


from .components import model as model
from .components import dataset as dataset
from .components.log import metrics
from .components.log import params
from .components.log import artifacts
from .components import auth as auth
from .components.auth import login
from .components.log import log
from .components.log import figure
from .components.log import predict
from .components.log import pip_requirement
from .components.log import resources

from .settings import set_backend, set_storage

from .evaluate import grader, eval

from .package import docker, fastapi

from .schema import Input, Output
from .predictor.predictor import BasePredictor


__version__ = "0.3.7"



from pydantic import BaseModel
from .paths import PathSchema
from .backend import BackendSchema
import os


class ModelSchema(BaseModel):
    paths: PathSchema = PathSchema().get_instance()
    backend: BackendSchema = BackendSchema().get_instance()

    PATH_MODEL_README: str = os.path.join(paths.PATH_MODEL_DIR, "ReadME.md")
    PATH_MODEL: str = os.path.join(paths.PATH_MODEL_DIR, "ran.pkl")



from sklearn.metrics import roc_auc_score


class ROC_AUC():

    def __init__(self):
        self.name = 'roc_auc'
        self.input_type = 'float'
        self.output_type = None
        self.kwargs = None
        

    def parse_data(self, data):
        
        return data



    def compute(self, references, predictions=None, prediction_scores=None, average="macro", sample_weight=None,
                max_fpr=None, multi_class="raise", labels=None, **kwargs):
        
        if prediction_scores is None and predictions is None:
            score = None
        elif predictions is None:
            score = roc_auc_score(y_true=references, y_score=prediction_scores, average=average, sample_weight=sample_weight,
                                multi_class=multi_class, max_fpr=max_fpr, labels=labels)
            score = float(score)
        elif prediction_scores is None:
            score = roc_auc_score(y_true=references, y_score=predictions, average=average, sample_weight=sample_weight,
                                multi_class=multi_class, max_fpr=max_fpr, labels=labels)
            score = float(score)
        
        score = {
            self.name : score
            }
 

        return score


from sklearn.metrics import confusion_matrix
import numpy as np


class ConfusionMatrix:
    def __init__(self):
        self.name = "confusion_matrix"
        self.input_type = "int"
        self.output_type = None
        self.kwargs = None

    def parse_data(self, data):

        return data

    def compute(
        self, references, predictions, normalize=None, sample_weight=None, **kwargs
    ):

        matrix = confusion_matrix(
            y_true=references,
            y_pred=predictions,
            normalize=normalize,
            sample_weight=sample_weight,
        )

        matrix = {self.name: np.array(matrix, dtype=float)}

        return matrix



from pureml.schema import BackendSchema


def set_backend(backend_url: str) -> None:
    """Set the backend url.

    Args:
        backend_url (str): The url of the backend.
    """
    backend: BackendSchema = BackendSchema().get_instance()
    backend.BASE_URL = backend_url



from pydantic import BaseModel, root_validator

from abc import ABC, abstractmethod

import typing
import pickle as pkl

from pydantic import BaseModel
import typing
import joblib

from pureml.packaging.errors import FrameworkNotSupportedError
from .model_framework import ModelFramework, ModelFrameworkType

# from . import MODEL_FRAMEWORKS_BY_TYPE, SUPPORTED_MODEL_FRAMEWORKS


from .packaging_utils import get_file_size
from .model_packaging.sklearn import SKLearn
from .model_packaging.catboost import CatBoost
from .model_packaging.xgboost import XGBoost
from .model_packaging.lightgbm import LightGBM
from .model_packaging.keras import Keras
from .model_packaging.tensorflow import Tensorflow
from .model_packaging.pytorch import Pytorch
from .model_packaging.pytorch_tabnet import PytorchTabnet
from .model_packaging.custom import Custom


MODEL_FRAMEWORKS_BY_TYPE = {
    ModelFrameworkType.SKLEARN: SKLearn(),
    ModelFrameworkType.XGBOOST: XGBoost(),
    ModelFrameworkType.LIGHTGBM: LightGBM(),
    ModelFrameworkType.CATBOOST: CatBoost(),
    ModelFrameworkType.KERAS: Keras(),
    # ModelFrameworkType.HUGGINGFACE_TRANSFORMER: HuggingfaceTransformer(),
    ModelFrameworkType.PYTORCH: Pytorch(),
    ModelFrameworkType.PYTORCH_TABNET: PytorchTabnet(),
    ModelFrameworkType.CUSTOM: Custom(),
}


SUPPORTED_MODEL_FRAMEWORKS = [
    ModelFrameworkType.SKLEARN,
    ModelFrameworkType.XGBOOST,
    ModelFrameworkType.LIGHTGBM,
    ModelFrameworkType.KERAS,
    ModelFrameworkType.TENSORFLOW,
    # ModelFrameworkType.HUGGINGFACE_TRANSFORMER,
    ModelFrameworkType.PYTORCH,
    ModelFrameworkType.PYTORCH_TABNET,
    ModelFrameworkType.CUSTOM,
]


class Model(ABC, BaseModel):
    model: typing.Any = None
    model_name: str = "model"
    model_path: str = None
    model_class: str = None

    model_config: dict = None
    model: typing.Any = None
    model_framework: typing.Any = None
    model_requirements: list = None

    # By default predict function of a framework should be assigned to here
    # If a user gives a predict function, assign it here
    predict: typing.Any = None

    @root_validator
    def set_fields(cls, values):

        return values

    # @staticmethod
    def from_dict(self):

        self.model = self.model_config["model"]
        # model_name = model_config_dict['model_name'],
        self.model_framework = self.model_config["model_framework"]
        self.model_requirements = self.model_config["model_requirements"]

    def generate_model_config(self):

        model_config = {
            "model": self.model,
            "model_framework": self.model_framework,
            "model_requirements": self.model_requirements,
            "model_size": get_file_size(pkl.dumps(self.model)),
        }

        return model_config

    def model_framework_from_model(self) -> ModelFramework:
        self.model_class = self.model.__class__
        model_framework = self.model_framework_from_model_class(self.model_class)

        return model_framework

    def model_framework_from_model_class(self, model_class) -> ModelFramework:

        for framework in MODEL_FRAMEWORKS_BY_TYPE.values():

            if framework.supports_model_class(model_class):
                return framework

        # raise FrameworkNotSupportedError(
        #     "Model must be one of "
        #     + "/".join([t.value for t in SUPPORTED_MODEL_FRAMEWORKS])
        # )
        framework = ModelFrameworkType.CUSTOM
        return framework

    def save_model(self):
        self.model_framework = self.model_framework_from_model()
        self.model_requirements = self.model_framework.get_requirements()

        # self.model_framework = ''
        # self.model_requirements = []

        self.model_config = self.generate_model_config()

        joblib.dump(self.model_config, self.model_path)

        return self.model_path

    def load_model(self):
        self.model_config = joblib.load(self.model_path)
        self.from_dict()

        return self.model



from pydantic import BaseModel
from .backend import BackendSchema

import typing
import os
from .paths import PathSchema


class FastAPISchema(BaseModel):
    paths: PathSchema = PathSchema().get_instance()
    PATH_FASTAPI_FILE = os.path.join(paths.PATH_PREDICT_DIR, "fastapi_server.py")
    PORT_FASTAPI = 8005  # Same port as docker server
    API_IP_HOST = "0.0.0.0"

    class Config:
        arbitrary_types_allowed = True


class DockerSchema(BaseModel):
    paths: PathSchema = PathSchema().get_instance()
    backend: BackendSchema = BackendSchema().get_instance()
    PATH_DOCKER_IMAGE = os.path.join(paths.PATH_PREDICT_DIR, "Dockerfile")
    PATH_DOCKER_CONFIG = os.path.join(paths.PATH_PREDICT_DIR, "DockerConfig.yaml")
    PORT_DOCKER = FastAPISchema().PORT_FASTAPI  # Same port as fastapi server
    PORT_HOST = 8000
    BASE_IMAGE_DOCKER = "python:3.8-slim"
    API_IP_DOCKER = FastAPISchema().API_IP_HOST
    API_IP_HOST = "0.0.0.0"

    class Config:
        arbitrary_types_allowed = True



import socket
from time import sleep, time
import requests
import typer
from rich import print
from pureml.cli.helpers import get_auth_headers, save_auth
from pureml.schema import ContentTypeHeader, AcceptHeader
from pureml.schema.backend import get_backend_base_url, get_frontend_base_url
from .orgs import select
from rich.progress import Progress, SpinnerColumn, TextColumn
from pureml.components import delete_token, get_api_token, get_org_id, get_token
from urllib.parse import urljoin
import json
import platform
import ipapi

app = typer.Typer()


def get_location():
    try:
        response = ipapi.location(output="json")
    except:
        try:
            # print("Getting device details...")
            response = requests.get(f"https://api64.ipify.org?format=json").json()
            response = requests.get(
                f"https://ipapi.co/{response['ip']}/json/",
                headers={"User-Agent": "pureml-cli"},
            ).json()
        except:
            response = {
                "ip": socket.gethostbyname(socket.gethostname()),
                "city": "Unknown",
                "region": "Unknown",
                "country": "Unknown",
            }
    location_data = {
        "ip": response["ip"] or socket.gethostbyname(socket.gethostname()),
        "city": response["city"] or "Unknown",
        "region": response["region"] or "Unknown",
        "country": response["country_name"] or "Unknown",
    }
    return location_data


@app.command()
def details():
    token = get_token()
    api_token = get_api_token()
    org_id = get_org_id()

    print("Org Id: ", org_id)
    if token is not None:
        print("Access Token: ", token)
    if api_token is not None:
        print("API Id: ", api_token["api_id"])
        print("API Key: ", api_token["api_key"])


@app.callback()
def callback():
    """
    Authentication user command

    Use with status, signup, login or logout option

    status - Checks current authenticated user
    signup - Creates new user
    login - Logs in user
    logout - Logs out user
    """


@app.command()
def signup(
    backend_url: str = typer.Option(
        "",
        "--backend-url",
        "-b",
        help="Backend URL for self-hosted or custom pureml backend instance",
    )
):
    try:
        print("\nCreate a new account/\n")
        email: str = typer.prompt("Enter new email")
        handle: str = typer.prompt("Enter new user handle")
        name: str = typer.prompt("Enter new user name")
        password: str = typer.prompt(
            "Enter new password", confirmation_prompt=True, hide_input=True
        )
        # organization_id: str = typer.prompt("Enter Organization id")
        # data = {"email": email, "password": password, "org": organization_id}
        data = {"email": email, "password": password, "handle": handle, "name": name}
        backend_base_url = get_backend_base_url(backend_url)

        url_path_1 = "user/signup"
        url = urljoin(backend_base_url, url_path_1)

        response = requests.post(url, json=data)

        if not response.ok:
            print(f"[red]Could not create account! Please try again later")
            return
        elif response.status_code == 202:
            print(
                f"[green]Successfully created your account! You need to verify your email to login!"
            )
            print(f"[green]You can then login using `pureml auth login`")
        else:
            print(f"[green]Successfully created your account!")
            print(f"[green]You can now login using `pureml auth login`")
    except Exception as e:
        print(f"[red]Could not create account! Please try again later")
        print(e)


@app.command()
def login(
    backend_url: str = typer.Option(
        "",
        "--backend-url",
        "-b",
        help="Backend URL for self-hosted or custom pureml backend instance",
    ),
    frontend_url: str = typer.Option(
        "",
        "--frontend-url",
        "-f",
        help="Frontend URL for self-hosted or custom pureml frontend instance",
    ),
    browserless: bool = typer.Option(
        False, "--browserless", "-s", help="Browserless login for ssh or pipelines"
    ),
    interactive: bool = typer.Option(
        False, "--interactive", "-i", help="Login with email and password interactively"
    ),
    api_key: bool = typer.Option(False, "--api-key", "-k", help="Login with api key"),
):
    try:
        backend_base_url = get_backend_base_url(backend_url)
        frontend_base_url = get_frontend_base_url(frontend_url)
        # Interactive login with email and password
        if interactive:
            print(f"\n[Enter your credentials to login[/\n")
            email: str = typer.prompt("Enter your email")
            password: str = typer.prompt("Enter your password", hide_input=True)
            data = {"email": email, "password": password}

            url_path = "user/login"
            url = urljoin(backend_base_url, url_path)

            response = requests.post(url, json=data)

            if response.ok:
                token = response.text
                token = json.loads(token)["data"][0]

                access_token = token["accessToken"]
                email = token["email"]

                save_auth(access_token=access_token, email=email)

                print(f"[green]Successfully logged in as {email}!")

                # Select organization
                select()

            else:
                print(f"[red]Unable to login to your account!")

        # API key login
        elif api_key:
            print(f"\n[Enter your credentials to login[/\n")
            api_id: str = typer.prompt("Enter your api id")
            api_key: str = typer.prompt("Enter your api key")
            save_auth(api_id=api_id, api_key=api_key)
            select()

        # Browser based login
        else:
            # Get device details
            device = platform.platform()
            device_data = get_location()
            device_id = device_data["ip"]
            device_location = (
                device_data["city"]
                + ", "
                + device_data["region"]
                + ", "
                + device_data["country"]
            )

            # Create session
            device_data = {
                "device": device,
                "device_id": device_id,
                "device_location": device_location,
            }
            url_path = "user/create-session"
            url = urljoin(backend_base_url, url_path)

            # print(device_data)
            response = requests.post(url, json=device_data)

            if not response.ok:
                print(f"[red]Unable to create session! Please try again later")
                return

            session_id = response.json()["data"][0]["session_id"]

            # Generater link & Open browser
            link = urljoin(frontend_base_url, f"verify-session/{session_id}")

            if browserless:
                print(
                    f"Please open the following link in your browser to login: [underline]{link}[/underline]"
                )
            else:
                # Open browser
                print(f"Opening the browser : [underline]{link}[/underline]")
                typer.launch(link)

            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
            ) as progress:
                progress.add_task(description="Waiting for response...", total=None)
                # Hit the endpoint to check if the session is verified
                start_time = time()
                while True:
                    url_path = "user/session-token"
                    url = urljoin(backend_base_url, url_path)
                    data = {
                        "session_id": session_id,
                        "device_id": device_id,
                    }
                    # print(data)
                    response = requests.post(url, json=data)
                    # print("Hit API response ", response.text)
                    if response.ok:
                        token = response.text
                        token = json.loads(token)["data"][0]

                        access_token = token["accessToken"]
                        email = token["email"]

                        save_auth(access_token=access_token, email=email)

                        break
                    else:
                        if response.status_code == 404:
                            print(f"[red]Session not found! Please try again later")
                            return
                        elif response.status_code == 403:
                            print(
                                f"[red]Session expired or invalid device! Please try again later"
                            )
                            return

                    if time() - start_time > 60 * 10:
                        print(f"[red]Session timed out!")
                        return

                    sleep(1)
            print(f"[green]Successfully logged in as {email}!")

            # Select organization
            select()
    except Exception as e:
        print(f"[red]Unable to contact the server! Please try again later")
        print(e)


@app.command()
def status():
    exists = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)
    if exists:
        print(f"[green]You are logged in!")


@app.command()
def logout():
    exists = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)
    if exists:
        delete_token()


if __name__ == "__main__":
    app()



from .metrics import MAE, MSE


class Regression():
    def __init__(self):
        self.task_type = 'regression'

        self.kwargs = None
        self.evaluator = None
        self.metrics = [MSE(), MAE()]

        self.scores = {}
    

    def compute(self):
        
        for m in self.metrics:
            #Adding  prediction scores to kwargs. It will be utilized my metrics needing it(roc_auc).
            self.kwargs['prediction_scores'] = self.prediction_scores

            score = m.compute(references=self.references, predictions=self.predictions, **self.kwargs)

            self.scores.update(score)

        return self.scores

    



import json
import os
from urllib.parse import urljoin
import time
import joblib
import pandas as pd
import requests
from pureml.cli.helpers import get_auth_headers

from pureml.schema import (
    DatasetSchema,
    StorageSchema,
    ConfigKeys,
    ContentTypeHeader,
    AcceptHeader,
)
from pureml.utils.hash import generate_hash_for_file
from pureml.utils.readme import load_readme
from rich import print

from . import get_org_id
from pureml.utils.version_utils import parse_version_label
from pureml.utils.config import reset_config

config_keys = ConfigKeys
storage = StorageSchema().get_instance()


def init_branch(label: str):
    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/{}/branch/create".format(org_id, name)
    url = urljoin(dataset_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_JSON)

    data = {"dataset_name": name, "branch_name": branch}

    data = json.dumps(data)

    response = requests.post(url, data=data, headers=headers)

    if response.ok:
        print(f"[bold green]Branch has been created!")

        return True

    else:
        print(f"[bold red]Branch has not been created!")

        return False


def check_dataset_hash(hash: str, label: str):

    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/{}/branch/{}/hash-status".format(org_id, name, branch)
    url = urljoin(dataset_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    data = {"hash": hash, "branch": branch}

    data = json.dumps(data)

    response = requests.post(url, data=data, headers=headers)

    hash_exists = False

    if response.ok:
        hash_exists = response.json()["data"][0]

    return hash_exists


def branch_details(label: str):
    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/{}/branch/{}".format(org_id, name, branch)
    url = urljoin(dataset_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        response_text = response.json()
        # T-1161 standardize api response to contains Data as a list
        details = response_text["data"]

        # details = response_text['data'][0]
        # print(details)

        return details

    else:
        print(f"[bold red]Branch details details have not been found")
        return


def branch_status(label: str):

    details = branch_details(label=label)

    if details:
        return True
    else:
        return False


def branch_delete(label: str) -> str:
    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/{}/branch/{}/delete".format(org_id, name, branch)
    url = urljoin(dataset_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.delete(url, headers=headers)

    if response.ok:
        print(f"[bold green]Dataset branch has been deleted")

    else:
        print(f"[bold red]Unable to delete Model branch")

    return response.text


def branch_list(label: str) -> str:

    name, _, _ = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/{}/branch".format(org_id, name)
    url = urljoin(dataset_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        response_text = response.json()
        branch_list = response_text["data"]

        return branch_list

    else:
        print(f"[bold red]Unable to obtain the list of branches!")

    return response.text


def list():
    """This function will return a list of all the datasets

    Returns
    -------
        A list of all the datasets

    """

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/all".format(org_id)
    url = urljoin(dataset_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print(f"[bold green]Obtained list of models")

        response_text = response.json()
        dataset_list = response_text["data"]
        # print(model_list)

        return dataset_list
    else:
        print(f"[bold red]Unable to obtain the list of dataset!")

    return


def init(label: str, readme: str = None):

    name, branch, _ = parse_version_label(label)

    headers = get_auth_headers(content_type=ContentTypeHeader.ALL)
    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    if readme is None:
        readme = dataset_schema.PATH_DATASET_README

    file_content, file_type = load_readme(path=readme)

    branch_user = "dev" if branch is None else branch
    branch_main = "main"

    url = "org/{}/dataset/{}/create".format(org_id, name)
    url = urljoin(dataset_schema.backend.BASE_URL, url)

    data = {
        "name": name,
        "branch_names": [branch_main, branch_user],
        "readme": {"file_type": file_type, "content": file_content},
    }

    data = json.dumps(data)
    files = {"file": (readme, open(readme, "rb"), file_type)}
    # response = requests.post(url, data=data, headers=headers, files=files)
    #
    response = requests.post(url, data=data, headers=headers)

    if response.ok:
        print(f"[bold green]Dataset has been created!")

        return True

    else:
        print(f"[bold red]Dataset has not been created!")

        return False


def save_dataset(dataset, name: str):
    dataset_schema = DatasetSchema()
    # file_name = '.'.join([name, 'parquet'])
    file_name = ".".join([name, "pkl"])
    save_path = os.path.join(dataset_schema.paths.PATH_DATASET_DIR, file_name)

    os.makedirs(dataset_schema.paths.PATH_DATASET_DIR, exist_ok=True)

    # dataset.to_parquet(save_path)
    joblib.dump(dataset, save_path)

    return save_path


def register(dataset, label: str, lineage, is_empty: bool = False) -> str:
    """The function takes in a dataset, a name and a version and saves the dataset locally, then uploads the
    dataset to the PureML server

    Parameters
    ----------
    dataset
        The dataset you want to register
    name : str
        The name of the dataset.
    version: str, optional
        The version of the dataset.

    """
    name, branch, _ = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    dataset_path = save_dataset(dataset, name)
    name_with_ext = dataset_path.split("/")[-1]

    dataset_hash = generate_hash_for_file(
        file_path=dataset_path, name=name, branch=branch, is_empty=is_empty
    )

    if is_empty:
        dataset_path = save_dataset(dataset=None, name=name)
        name_with_ext = dataset_path.split("/")[-1]

    dataset_exists = dataset_status(label)
    # print('Dataset status', dataset_exists)

    if not dataset_exists:
        dataset_created = init(label=label)
        # print('dataset_created', dataset_created)
        if not dataset_created:
            print("[bold red] Unable to register the dataset")
            return False, dataset_hash, None
    else:
        print("[bold green] Connected to Dataset")

    branch_exists = branch_status(label=label)
    # print('Branch status', branch_exists)

    if not branch_exists:
        branch_created = init_branch(label=label)
        # print('branch_created', branch_created)

        if not branch_created:
            print("[bold red] Unable to register the dataset")
            return False, dataset_hash, None
    else:
        print("[bold green] Connected to Branch")

    dataset_exists_remote = check_dataset_hash(hash=dataset_hash, label=label)

    if dataset_exists_remote:

        print(f"[bold red]Dataset already exists. Not registering a new version!")
        return True, dataset_hash, "latest"
    else:

        url = "org/{}/dataset/{}/branch/{}/register".format(org_id, name, branch)
        url = urljoin(dataset_schema.backend.BASE_URL, url)

        headers = get_auth_headers(content_type=None, accept=AcceptHeader.APP_JSON)

        files = {"file": (name_with_ext, open(dataset_path, "rb"))}

        lineage = json.dumps(lineage)

        data = {
            "name": name,
            "branch": branch,
            "hash": dataset_hash,
            "lineage": lineage,
            "is_empty": is_empty,
            "storage": storage.STORAGE,
        }

        # data = json.dumps(data)

        response = requests.post(url, files=files, data=data, headers=headers)
        # print(response.request.headers)
        # print(response.request.body)

        if response.ok:

            dataset_version = None

            # print(response.json())
            try:
                dataset_version = response.json()["data"][0]["version"]
                print("Version: ", dataset_version)
                dataset_label = ":".join([label, dataset_version])
                print("Dataset label: ", dataset_label)

                if is_empty:
                    print(f"[bold green]Lineage has been registered!")
                else:
                    print(f"[bold green]Dataset and lineage have been registered!")

                # reset_config(key=config_keys.dataset.value)

            except Exception as e:
                print(
                    "[bold red] Incorrect json response. Dataset has not been registered"
                )
                print(e)

            return True, dataset_hash, dataset_version
        else:
            print(f"[bold red]Dataset has not been registered!")
            print(response.text)

            return True, dataset_hash, None


def dataset_status(label: str):

    name, _, _ = parse_version_label(label)

    dataset_details = details(label=label)

    if dataset_details:
        return True
    else:
        return False


def details(
    label: str,
):
    """It fetches the details of a dataset.

    Parameters
    ----------
    name : str
        The name of the dataset
    version: str
        The version of the dataset
    Returns
    -------
        The details of the dataset.

    """

    name, _, _ = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/{}".format(org_id, name)
    url = urljoin(dataset_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print(f"[bold green]Dataset details have been fetched")
        response_text = response.json()
        dataset_details = response_text["data"][0]

        return dataset_details

    else:
        print(f"[bold red]Dataset details have not been found")
        return


def version_details(label: str):
    """It fetches the details of a dataset.

    Parameters
    ----------
    name : str
        The name of the dataset
    version: str
        The version of the dataset
    Returns
    -------
        The details of the dataset.

    """

    name, branch, version = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/{}/branch/{}/version/{}".format(org_id, name, branch, version)
    url = urljoin(dataset_schema.backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print(f"[bold green]Dataset Version details have been fetched")
        response_text = response.json()
        dataset_details = response_text["data"][0]
        # print(dataset_details)

        return dataset_details

    else:
        print(f"[bold red]Dataset details have not been found")
        return


def fetch(label: str):
    """This function fetches a dataset from the server and returns it as a dataframe object

    Parameters
    ----------
    name : str, optional
        The name of the dataset you want to fetch.
    version: str
        The version of the dataset

    Returns
    -------
        The dataset dataframe is being returned.

    """

    name, branch, version = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    dataset_details = version_details(label=label)

    if dataset_details is None:
        print(f"[bold red]Unable to fetch Dataset version")
        return

    is_empty = dataset_details["is_empty"]

    if is_empty:
        print("[bold orange]Dataset file is not registered to the version")
        return

    # storage_path = dataset_details["path"]["source_path"]
    # storage_source_type = dataset_details["path"]["source_type"]["public_url"]

    # dataset_url = urljoin(storage_source_type, storage_path)

    dataset_url = dataset_details["path"]

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    # print('url', dataset_url)

    response = requests.get(dataset_url)

    if response.ok:
        dataset_bytes = response.content
        # open('temp_dataset.parquet', 'wb').write(dataset_bytes)
        # dataset = pd.read_parquet('temp_dataset.parquet')

        open("temp_dataset.pkl", "wb").write(dataset_bytes)
        dataset = joblib.load("temp_dataset.pkl")

        # print(f"[bold green]Dataset has been fetched")
        return dataset
    else:
        print(f"[bold red]Unable to fetch Dataset")
        # print(response.status_code)
        # print(response.text)
        # print(response.url)
        return


# def delete(label: str) -> str:
#     """This function deletes a dataset from the project

#     Parameters
#     ----------
#     name : str
#         The name of the dataset you want to delete
#     version : str
#         The version of the dataset to delete.

#     """

#     name, branch, version = parse_version_label(label)

#     user_token = get_token()
#     org_id = get_org_id()
#     dataset_schema = DatasetSchema()

#     url = "org/{}/dataset/{}/delete".format(org_id, name)
#     url = urljoin(dataset_schema.backend.BASE_URL, url)

#     headers = {
#         "Content-Type": ContentTypeHeader.APP_FORM_URL_ENCODED.value,
#         "Authorization": "Bearer {}".format(user_token),
#     }

#     data = {"version": version}

#     data = json.dumps(data)

#     response = requests.delete(url, headers=headers, data=data)

#     if response.ok:
#         print(f"[bold green]Dataset has been deleted")

#     else:
#         print(f"[bold red]Unable to delete Dataset")

#     return response.text



from pathlib import Path
from typing import Optional
import jwt
import requests

# import typer
from rich import print
from rich.syntax import Syntax

import os
import json
import typing

from urllib.parse import urljoin

from pureml.cli.helpers import get_auth_headers

from . import get_token, get_org_id

from pureml.schema import PathSchema, BackendSchema, ContentTypeHeader
from joblib import Parallel, delayed

path_schema = PathSchema().get_instance()
backend_schema = BackendSchema().get_instance()


def details(
    model_name: str, model_branch: str, model_version: str = "latest", name: str = ""
):
    """This function returns the details of the audio for a given model

    Parameters
    ----------
    model_name : str
        The name of the model you want to get the audio details for
    model_version: str
        The version of the model
    name : str
        The name of the audio.

    """

    org_id = get_org_id()

    url = "org/{}/model/{}/branch/{}/version/{}/log".format(
        org_id, model_name, model_branch, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.status_code == 200:
        res_text = json.loads(response.text)

        if len(res_text) == 0:
            print("[bold yellow] No audio have been found for the model")
            print(res_text)
            return
        else:
            print("[bold green]audio have been found for the model")
            print(res_text)
            return res_text

    else:
        print("[bold red]Unable to obtain the audio details")
        print(response.text)
        return


def add(
    audio: str, model_name: str, model_branch: str, model_version: str = "latest"
) -> str:
    """`add` function takes in the path of the audio, name of the audio and the model name and
    registers the audio

    Parameters
    ----------
    audio : str
        The path to the audio file.
    name : str
        The name of the audio.
    model_name : str
        The name of the model you want to add audio to.
    model_version: str
        The version of the model

    Returns
    -------
        The response is a JSON object

    """

    org_id = get_org_id()

    url = "org/{}/model/{}/branch/{}/version/{}/log".format(
        org_id, model_name, model_branch, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    user_token = get_token()

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    files = {}
    for file_name, file_path in audio.items():

        if os.path.isfile(file_path):
            files[file_name] = open(file_path, "rb")
        else:
            print("[bold red] audio", file_name, "does not exist at the given path")

    data = {"name_path_mapping": audio}

    response = requests.post(url, data=data, files=files, headers=headers)

    if response.status_code == 200:
        print(f"[bold green]audios have been registered!")

    else:
        print(f"[bold red]audios have not been registered!")
        print(response.text)

    return response.text


def fetch(
    model_name: str, model_branch: str, model_version: str = "latest", name: str = ""
):
    """It fetches the audio from the server and stores it in the local directory

    Parameters
    ----------
    model_name : str
        The name of the model you want to fetch the audio from.
    model_version: str
        The version of the model
    name : str
        The name of the audio to be fetched. If not specified, all audios will be fetched.

    Returns
    -------
        The response text is being returned.

    """

    org_id = get_org_id()

    def fetch_audio(audio_details: dict):

        url = audio_details["location"]
        file_path_temp = audio_details["path"]
        file_name = file_path_temp.split(os.path.sep)[-1]
        save_path = os.path.join(path_schema.PATH_AUDIO_DIR, file_name)
        print("save path", save_path)

        name_fetched = audio_details["audio"]

        headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

        print("audio url", url)

        # response = requests.get(url, headers=headers)
        response = requests.get(url)

        print(response.status_code)

        if response.status_code == 200:
            print("[bold green] audio {} has been fetched".format(name_fetched))

            save_dir = os.path.dirname(save_path)

            os.makedirs(save_dir, exist_ok=True)

            audio_bytes = response.content

            open(save_path, "wb").write(audio_bytes)

            print(
                "[bold green] audio {} has been stored at {}".format(
                    name_fetched, save_path
                )
            )

            return response.text
        else:
            print("[bold red] Unable to fetch the audio")

            return response.text

    audio_details = details(
        model_name=model_name,
        model_branch=model_branch,
        name=name,
        model_version=model_version,
    )

    if audio_details is None:
        return

    if type(audio_details) is dict:

        res_text = fetch_audio(audio_details)

    elif type(audio_details) is list:
        res_text = Parallel(n_jobs=-1)(
            delayed(fetch_audio)(art_det) for art_det in audio_details
        )

    return res_text


# def delete(
#     name: str, model_name: str, model_branch: str, model_version: str = "latest"
# ) -> str:
#     """`delete()` deletes an audio from a model

#     Parameters
#     ----------
#     name : str
#         The name of the audio you want to delete.
#     model_name : str
#         The name of the model you want to delete the audio from
#     model_version: str
#         The version of the model

#     """

#     user_token = get_token()
#     org_id = get_org_id()

#     url = "org/{}/model/{}/branch/{}/version/{}/log".format(
#         org_id, model_name, model_branch, model_version
#     )
#     url = urljoin(backend_schema.BASE_URL, url)

#     headers = {
#         "Content-Type": "application/x-www-form-urlencoded",
#         "Authorization": "Bearer {}".format(user_token),
#     }

#     # audio_details = details(model_name=model_name, audio=audio)

#     # if audio_details is None:
#     #     print('[bold red] Unable to find audio details')
#     #     return

#     response = requests.delete(url, headers=headers)

#     if response.status_code == 200:
#         print(f"[bold green]audio has been deleted")

#     else:
#         print(f"[bold red]Unable to delete audio")

#     return response.text



from pureml.utils.pipeline import add_transformer_to_config
import functools


def transformer(parent: str = None):
    def decorator(func):
        # print('Inside decorator')
        # print("decorating", func, "with argument", name)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # print("Inside wrapper")
            func_name = func.__name__
            func_description = func.__doc__

            # print("func_name", func_name)
            # print("func_description", func_description)

            func_output = func(*args, **kwargs)

            add_transformer_to_config(
                name=func_name, description=func_description, func=func, parent=parent
            )

            res_text = ""

            return func_output

        # print("Outside  wrapper")

        return wrapper

    # print('Outside decorator')

    return decorator



import os
import shutil
import uvicorn
from fastapi import FastAPI
from pureml.schema import FastAPISchema, PredictSchema, PathSchema
from pureml.utils.package import process_input, process_output
from pureml.utils.version_utils import parse_version_label
from pureml import predict, pip_requirement, resources
from pureml.utils.resources import zip_content, unzip_content
import sys


prediction_schema = PredictSchema()
fastapi_schema = FastAPISchema()
path_schema = PathSchema()


def get_resources(label, resources_path):
    resources.fetch(label=label)

    if not os.path.exists(prediction_schema.PATH_RESOURCES):

        if resources_path is None:

            resources_path = prediction_schema.PATH_RESOURCES_DIR_DEFAULT
            print("Taking the default resources path: ", resources_path)
        else:
            print("Taking the resources path: ", resources_path)

        zip_content(resources_path, prediction_schema.PATH_RESOURCES)

        if os.path.exists(prediction_schema.PATH_RESOURCES):
            unzip_content(
                prediction_schema.PATH_RESOURCES, path_schema.PATH_PREDICT_DIR
            )
        else:
            raise Exception(resources_path, "does not exists!!!")
    else:
        print("Taking the fetched resources file path")


def get_predict_file(label, predict_path):

    predict.fetch(label)

    if not os.path.exists(prediction_schema.PATH_PREDICT):
        print("[orange] Prediction Function is not logged")

        # os.makedirs(PATH_PREDICT_DIR, exist_ok=True)

        if predict_path is None:
            predict_path = prediction_schema.PATH_PREDICT_USER
            print("Taking the default predict.py file path: ", predict_path)
        else:
            print("Taking the predict.py file path: ", predict_path)

        if os.path.exists(predict_path):
            shutil.copy(predict_path, prediction_schema.PATH_PREDICT)
        else:
            raise Exception(predict_path, "does not exists!!!")
    else:
        print("Taking the fetched predict.py file path")


def get_requirements_file(label, requirements_path):
    pip_requirement.fetch(label=label)

    if not os.path.exists(prediction_schema.PATH_PREDICT_REQUIREMENTS):

        if requirements_path is None:
            requirements_path = prediction_schema.PATH_PREDICT_REQUIREMENTS_USER
            print("Taking the default requirements.txt file path: ", requirements_path)
        else:
            print("Taking the requirements.txt file path: ", requirements_path)

        if os.path.exists(requirements_path):
            shutil.copy(requirements_path, prediction_schema.PATH_PREDICT_REQUIREMENTS)
        else:
            raise Exception(requirements_path, "does not exists!!!")
    else:
        print("Taking the fetched requirements.txt file path")


def generate_api(label: str):
    name, branch, version = parse_version_label(label=label)

    api = """
from fastapi import FastAPI, Depends, Request, UploadFile, File
from fastapi.middleware.cors import CORSMiddleware
import uvicorn
import pureml
from predict import Predictor
import os
from dotenv import load_dotenv
import json
import shutil
from pureml.utils.package import process_input, process_output
from typing import Union, Optional
from pureml.utils.predict import predict_request_with_json, predict_request_with_file

import nest_asyncio
from pyngrok import ngrok

load_dotenv()

org_id = os.getenv('ORG_ID')
api_id = os.getenv('API_ID')
api_key = os.getenv('API_KEY')

if api_id is not None and api_key is not None and org_id is not None:
    pureml.login(org_id=org_id, api_id=api_id, api_key=api_key)

predictor = Predictor()
predictor.load_models()

# input_type, input_shape = process_input(input=predictor.input)
# output_type, output_shape = process_output(output=predictor.output)


# Create the app
app = FastAPI()     

# middlewares
app.add_middleware(
    CORSMiddleware,
    allow_origins=['*'],
    allow_credentials=True, 
    allow_methods=['*'], 
    allow_headers=['*']
)


@app.post('/predict')
async def predict(request:Request = None, file: Optional[UploadFile] = File(None)):
    if request is None and file is None:
        print('Error in data input format')
        predictions = json.dumps(None)
        return predictions
    

    if request is not None and file is None:
        print("Processing request")
        predictions = await predict_request_with_json(request=request, predictor=predictor)


    if file is not None:
        print("Processing uploaded file")
        predictions = await predict_request_with_file(file=file, predictor=predictor)
        
    return predictions

if __name__ == '__main__':
    ngrok_tunnel = ngrok.connect({PORT})

    # Public URL for fastapi server
    print('Public URL:', ngrok_tunnel.public_url)

    nest_asyncio.apply()

    uvicorn.run(app, host='{HOST}', port={PORT})""".format(
        HOST=fastapi_schema.API_IP_HOST, PORT=fastapi_schema.PORT_FASTAPI
    )

    return api


def create_fastapi_file(
    label, predict_path=None, requirements_path=None, resources_path=None
):
    fastapi_schema = FastAPISchema()

    get_resources(label, resources_path)

    get_predict_file(label, predict_path)

    get_requirements_file(label, requirements_path)

    api = generate_api(label=label)

    with open(fastapi_schema.PATH_FASTAPI_FILE, "w") as api_writer:
        api_writer.write(api)

    api_writer.close()

    print("FastAPI server files are created")


#     print("""
#           API sucessfully created. To run your API, please run the following command
# --> !python <api_name>
#           """)


def run(label, predict_path=None, requirements_path=None):

    create_fastapi_file(
        label, predict_path=predict_path, requirements_path=requirements_path
    )

    interpreter_path = str(sys.executable)

    if os.path.exists(interpreter_path):

        run_command = "{interpreter_path} '{api_path}'".format(
            interpreter_path=interpreter_path, api_path=fastapi_schema.PATH_FASTAPI_FILE
        )

        os.system(run_command)
    else:
        print("Interpreter not found at", interpreter_path)

    # app = FastAPI()

    # uvicorn.run(app, host=fastapi_schema.API_IP_HOST, port=fastapi_schema.PORT_FASTAPI)



import os
from pathlib import Path
from .packaging import Model


def save_model(model, model_name, model_path) -> None:
    """The function takes in a model and a model name, and saves the model to the default `models` directory.

    Parameters
    ----------
    model
        The model you want to save.
    model_name
        The name of the model.

    """
    model_dir = os.path.dirname(model_path)
    # print(model_dir)
    # os.makedirs(model_dir, exist_ok=True)

    Model(model=model, model_name=model_name, model_path=model_path).save_model()


def load_model(model_path: Path):
    """Loads a model from a given path

    Parameters
    ----------
    model_path : Path
        The path to the model you want to load.

    Returns
    -------
        The model is being returned.

    """
    model = Model(model_path=model_path).load_model()

    return model



from .dataset import DatasetSchema
from .paths import PathSchema
from .model import ModelSchema
from .predict import PredictSchema, Input, Output, DataTypes
from .storage import StorageSchema
from .packaging import DockerSchema, FastAPISchema
from .log import LogSchema
from .backend import BackendSchema
from .types import DataTypes
from .config import ConfigKeys
from .request import ContentTypeHeader, AcceptHeader
from .env import Env



from pydantic import BaseModel
from enum import Enum


class ContentTypeHeader(Enum):
    APP_FORM_URL_ENCODED = "application/x-www-form-urlencoded"
    APP_JSON = "application/json"
    ALL = "*/*"
    MULTIPART_FORMDATA = "multipart/form-data"


class AcceptHeader(Enum):
    APP_JSON = "application/json"



from sklearn.metrics import mean_absolute_error


class MAE():

    def __init__(self):
        self.name = 'mae'
        self.input_type = 'float'
        self.output_type = None
        self.kwargs = None
        

    def parse_data(self, data):
        
        return data



    def compute(self, predictions, references, sample_weight=None, multioutput="uniform_average", **kwargs):
        
        score = mean_absolute_error(y_true=references, y_pred=predictions, sample_weight=sample_weight, multioutput=multioutput)
        
        score = {
            self.name : float(score)
            }
 

        return score


import json
import os
from urllib.parse import urljoin

import numpy as np
import requests
from joblib import Parallel, delayed
from pureml.cli.helpers import get_auth_headers

from pureml.utils.pipeline import add_pred_to_config
from pureml.schema import (
    PathSchema,
    BackendSchema,
    StorageSchema,
    LogSchema,
    ConfigKeys,
    AcceptHeader,
    ContentTypeHeader,
)
from rich import print
from . import get_org_id, pip_requirement, resources
import shutil
from pureml.utils.version_utils import parse_version_label
from pureml.utils.config import reset_config


path_schema = PathSchema().get_instance()
backend_schema = BackendSchema().get_instance()
post_key_predict = LogSchema().key.predict.value
post_key_requirements = LogSchema().key.requirements.value
post_key_resources = LogSchema().key.resources.value
config_keys = ConfigKeys
storage = StorageSchema().get_instance()


def post_predict(file_paths, model_name: str, model_branch: str, model_version: str):
    org_id = get_org_id()

    url = "org/{}/model/{}/branch/{}/version/{}/logfile".format(
        org_id, model_name, model_branch, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=None, accept=AcceptHeader.APP_JSON)

    files = []
    for file_name, file_path in file_paths.items():
        # print("filename", file_name)

        if os.path.isfile(file_path):
            files.append(("file", (file_name, open(file_path, "rb"))))

        else:
            print("[bold red] Predict", file_name, "does not exist at the given path")

    data = {
        "data": file_paths,
        "key": post_key_predict,
        "storage": storage.STORAGE,
    }

    # data = json.dumps(data)

    response = requests.post(url, data=data, files=files, headers=headers)

    if response.ok:
        print(f"[bold green]Predict Function has been registered!")
        reset_config(key=config_keys.pred_function.value)

    else:
        print(f"[bold red]Predict Function has not been registered!")

    return response


def add(label: str = None, paths: dict = None) -> str:

    model_name, model_branch, model_version = parse_version_label(label)

    pred_path = paths[post_key_predict]

    if post_key_requirements in paths.keys():
        pip_requirement.add(label, path=paths[post_key_requirements])

    if post_key_resources in paths.keys():
        resources.add(label, path=paths[post_key_resources])

    file_paths = {post_key_predict: pred_path}

    add_pred_to_config(
        values=pred_path,
        model_name=model_name,
        model_branch=model_branch,
        model_version=model_version,
    )

    if (
        model_name is not None
        and model_version is not None
        and model_version is not None
    ):
        response = post_predict(
            file_paths=file_paths,
            model_name=model_name,
            model_branch=model_branch,
            model_version=model_version,
        )

        print(response.text)

    # return response.text


def details(label: str):
    model_name, model_branch, model_version = parse_version_label(label)
    org_id = get_org_id()

    url = "org/{}/model/{}/branch/{}/version/{}/log".format(
        org_id, model_name, model_branch, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=None, accept=AcceptHeader.APP_JSON)

    response = requests.get(url, headers=headers)

    if response.ok:
        # T-1161 standardize api response to contains Models as a list
        response_text = response.json()
        details = response_text["data"]

        # print(model_details)

        return details

    else:
        print(f"[bold red]Unable to fetch Predict!")
        return


def fetch(label: str):
    model_name, model_branch, model_version = parse_version_label(label)

    org_id = get_org_id()

    def fetch_predict(file_details):

        file_name, url = file_details

        save_path = os.path.join(path_schema.PATH_PREDICT_DIR, file_name)
        # print("save path", save_path)

        headers = get_auth_headers(
            content_type=ContentTypeHeader.APP_FORM_URL_ENCODED,
            accept=AcceptHeader.APP_JSON,
        )
        # print("predict url", url)

        response = requests.get(url)

        # print(response.status_code)

        if response.ok:
            print("[bold green] predict file {} has been fetched".format(file_name))

            save_dir = os.path.dirname(save_path)

            os.makedirs(save_dir, exist_ok=True)

            predict_bytes = response.content

            open(save_path, "wb").write(predict_bytes)

            # print(
            #     "[bold green] predict file {} has been stored at {}".format(
            #         file_name, save_path
            #     )
            # )

            return response.text
        else:
            print("[bold red] Unable to fetch the predict function")

            return response.text

    predict_details = details(label=label)

    if predict_details is None:
        return

    # pred_urls = give_predict_urls(details=predict_details)
    pred_urls = give_predict_url(details=predict_details, key=post_key_predict)

    if len(pred_urls) == 1:

        res_text = fetch_predict(pred_urls[0])

    else:
        res_text = Parallel(n_jobs=-1)(
            delayed(fetch_predict)(pred_url) for pred_url in pred_urls
        )


def give_predict_url(details, key: str):
    predict_paths = []
    # file_url = None
    source_path = None
    file_url = None
    # print(details)

    if details is not None:

        for det in details:
            # print(det["key"])
            if det["key"] == key:
                source_path = det["key"]
                file_url = det["data"]
                source_path = ".".join([source_path, "py"])
                # source_path = os.path.join(path_schema.PATH_PREDICT_DIR, source_path)
                predict_paths.append([source_path, file_url])

                # print(source_path, file_url)

                return predict_paths
    print("[bold red] Unable to find the predict file")

    return predict_paths



import shutil
import zipfile
from pureml.schema import PredictSchema
import os


# def zip_content(src_path, dst_path):
#     print(src_path)
#     ignore_patterns = shutil.ignore_patterns(".pureml", "*.venv", "*.pyc")

#     # Create a zip archive of the folder
#     shutil.make_archive(
#         base_dir=src_path,
#         root_dir=src_path,
#         format=PredictSchema().resource_format,
#         base_name=dst_path,
#         ignore=ignore_patterns,
#     )


def remove_dirs(dirs):
    dirs_to_ignore = PredictSchema().dirs_to_ignore

    for d in dirs_to_ignore:
        if d in dirs:
            # print('Removing ', d)
            dirs.remove(d)
    # print(dirs)

    return dirs


def zip_content(src_path, dst_path):

    # [".pureml", ".venv", "__pycache__"]
    dirs_to_ignore = PredictSchema().dirs_to_ignore

    # Create a zip archive of the folder, excluding the specified folder
    with zipfile.ZipFile(dst_path, "w", zipfile.ZIP_DEFLATED) as zip_file:

        for root, dirs, files in os.walk(src_path):
            # print("dirs", dirs)
            dirs = remove_dirs(dirs)

            for file in files:
                file_path = os.path.join(root, file)
                # print("outside for loop", file_path)
                if not any(file_path.startswith(folder) for folder in dirs_to_ignore):
                    zip_file.write(file_path, file_path)
                    print(file_path)

    print("Resources are collected")


def unzip_content(src_path, dst_path):
    shutil.unpack_archive(src_path, dst_path, format=PredictSchema().resource_format)

    # Delete the zip file
    # os.remove(src_path)



from pydantic import BaseModel
from .paths import PathSchema
import os
from .types import DataTypes


class Input(BaseModel):
    type: DataTypes
    shape: tuple = None


class Output(BaseModel):
    type: DataTypes = None
    shape: tuple = None


class PredictSchema(BaseModel):
    paths: PathSchema = PathSchema().get_instance()

    PREDICT_NAME: str = "predict.py"
    REQUIREMENTS_NAME: str = "requirements.txt"
    RESOURCES_NAME: str = "resources.zip"

    PATH_PREDICT: str = os.path.join(paths.PATH_PREDICT_DIR, PREDICT_NAME)
    PATH_PREDICT_USER: str = os.path.join(os.getcwd(), PREDICT_NAME)

    PATH_PREDICT_REQUIREMENTS: str = os.path.join(
        paths.PATH_PREDICT_DIR, REQUIREMENTS_NAME
    )
    PATH_PREDICT_REQUIREMENTS_USER: str = os.path.join(os.getcwd(), REQUIREMENTS_NAME)

    PATH_RESOURCES: str = os.path.join(paths.PATH_PREDICT_DIR, RESOURCES_NAME)
    PATH_RESOURCES_DIR_DEFAULT: str = "./"  # os.getcwd()

    resource_format: str = "zip"
    dirs_to_ignore: list = [".pureml", ".venv", "__pycache__"]

    class Config:
        arbitrary_types_allowed = True



from .metrics import Accuracy, Precision, Recall, F1, ConfusionMatrix


class Classification:
    def __init__(self):
        self.task_type = "classification"

        self.kwargs = None

        self.references = None
        self.predictions = None
        self.prediction_scores = None

        self.label_type = "binary"

        self.metrics = [
            Accuracy(),
            Precision(),
            Recall(),
            F1(),
            ConfusionMatrix(),
        ]  # , ROC_AUC()]
        self.scores = {}

    def compute(self):
        self.setup()

        for m in self.metrics:
            # Adding  prediction scores to kwargs. It will be utilized my metrics needing it(roc_auc).
            self.kwargs["prediction_scores"] = self.prediction_scores

            score = m.compute(
                references=self.references, predictions=self.predictions, **self.kwargs
            )

            self.scores.update(score)

        return self.scores

    def setup(self):
        self.is_multiclass()
        self.setup_kwargs()

    def get_predictions(self):
        pass

    def is_multiclass(self):
        if self.predictions is not None:
            labels_all = set(self.references).union(self.predictions)
            if len(labels_all) > 2:
                self.label_type = "multilabel"

    def setup_kwargs(self):
        if "average" not in self.kwargs:
            if self.label_type == "multilabel":
                self.kwargs["average"] = "micro"



import json
import os
from urllib.parse import urljoin

import numpy as np
import requests
from joblib import Parallel, delayed
from PIL import Image
from pureml.cli.helpers import get_auth_headers

from pureml.utils.pipeline import add_figures_to_config
from pureml.schema import (
    PathSchema,
    BackendSchema,
    StorageSchema,
    LogSchema,
    ConfigKeys,
    AcceptHeader,
    ContentTypeHeader,
)
from rich import print
from . import get_org_id

from pureml.utils.version_utils import parse_version_label
from pureml.utils.config import reset_config


path_schema = PathSchema().get_instance()
backend_schema = BackendSchema().get_instance()
post_key_figure = LogSchema().key.figure.value
config_keys = ConfigKeys
storage = StorageSchema().get_instance()


def save_images(figure):

    os.makedirs(path_schema.PATH_FIGURE_DIR, exist_ok=True)
    figure_paths = {}
    for figure_key, figure_value in figure.items():
        if figure_key.rsplit(".", 1)[-1] == "png":
            save_name = figure_key
        else:
            save_name = os.path.join(
                path_schema.PATH_FIGURE_DIR, ".".join([figure_key, "png"])
            )

        canvas = figure_value.canvas
        canvas.draw()
        data = np.frombuffer(canvas.tostring_rgb(), dtype=np.uint8)
        rgb_array = data.reshape(canvas.get_width_height()[::-1] + (3,))

        data = Image.fromarray(rgb_array)
        data.save(save_name)

        figure_paths[figure_key] = save_name

    return figure_paths


def post_figures(figure_paths, model_name: str, model_branch: str, model_version: str):
    org_id = get_org_id()

    # print('figure_paths', figure_paths)

    url = "org/{}/model/{}/branch/{}/version/{}/logfile".format(
        org_id, model_name, model_branch, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=None, accept=AcceptHeader.APP_JSON)

    files = []
    for file_name, file_path in figure_paths.items():
        # print("filename", file_name)

        if os.path.isfile(file_path):
            files.append(("file", (file_name, open(file_path, "rb"))))

        else:
            print("[bold red] figure", file_name, "does not exist at the given path")

    data = {
        "data": figure_paths,
        "key": "figure",
        "storage": storage.STORAGE,
    }

    # data = json.dumps(data)

    # try:

    response = requests.post(url, data=data, files=files, headers=headers)

    if response.ok:
        print(f"[bold green]Figures have been registered!")
        reset_config(key=config_keys.figure.value)

    else:
        print(f"[bold red]Figures have not been registered!")
        print(response.text)

    return response
    # except Exception as e:
    #     return


def add(label: str = None, figure: dict = None, file_paths: dict = None) -> str:
    """`add` function takes in the path of the figure, name of the figure and the model name and
    registers the figure

    Parameters
    ----------
    figure : dict
        Key is the figure name, value is the matplotlib figure object
    name : str
        The name of the figure.
    model_name : str
        The name of the model you want to add figure to.
    model_version: str
        The version of the model

    Returns
    -------
        The response is a JSON object

    """
    model_name, model_branch, model_version = parse_version_label(label)
    # print(model_name, model_branch, model_version)
    # print('file_paths', file_paths)
    # print('figure', figure)

    if file_paths is None:
        file_paths = save_images(figure)
        # print('figre paths', figure_paths)

    add_figures_to_config(
        values=file_paths,
        model_name=model_name,
        model_branch=model_branch,
        model_version=model_version,
    )

    if (
        model_name is not None
        and model_version is not None
        and model_version is not None
    ):
        response = post_figures(
            figure_paths=file_paths,
            model_name=model_name,
            model_branch=model_branch,
            model_version=model_version,
        )

        # print(response.text)

    # return response.text


def details(label: str):
    model_name, model_branch, model_version = parse_version_label(label)
    org_id = get_org_id()

    url = "org/{}/model/{}/branch/{}/version/{}/log".format(
        org_id, model_name, model_branch, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # T-1161 standardize api response to contains Models as a list
        response_text = response.json()
        details = response_text["data"]

        # print(model_details)

        return details

    else:
        print(f"[bold red]Unable to fetch Figure!")
        return


def fetch(label: str, key: str):
    """It fetches the figure from the server and stores it in the local directory

    Parameters
    ----------
    model_name : str
        The name of the model you want to fetch the figure from.
    model_version: str
        The version of the model
    name : str
        The name of the figure to be fetched. If not specified, all figures will be fetched.

    Returns
    -------
        The response text is being returned.

    """
    model_name, model_branch, model_version = parse_version_label(label)

    org_id = get_org_id()

    def fetch_figure(file_details):

        file_name, url = file_details

        save_path = os.path.join(path_schema.PATH_FIGURE_DIR, file_name)
        # print("save path in fetching", save_path)

        headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

        # print("figure url", url)

        # response = requests.get(url, headers=headers)
        response = requests.get(url)

        # print(response.status_code)

        if response.ok:
            print("[bold green] figure {} has been fetched".format(file_name))

            save_dir = os.path.dirname(save_path)

            os.makedirs(save_dir, exist_ok=True)

            figure_bytes = response.content

            open(save_path, "wb").write(figure_bytes)

            print(
                "[bold green] figure {} has been stored at {}".format(
                    file_name, save_path
                )
            )

            return response.text
        else:
            print("[bold red] Unable to fetch the figure")

            return response.text

    figure_details = details(label=label)

    if figure_details is None:
        return

    # fig_urls = give_fig_urls(details=figure_details)
    fig_urls = give_fig_url(details=figure_details, key=key)

    if len(fig_urls) == 1:

        res_text = fetch_figure(fig_urls[0])

    else:
        res_text = Parallel(n_jobs=-1)(
            delayed(fetch_figure)(fig_url) for fig_url in fig_urls
        )

    # return res_text


def give_fig_url(details, key: str):
    fig_paths = []
    # file_url = None
    source_path = None
    file_url = None
    # print(details)

    if details is not None:

        for det in details:
            # print(det["key"])
            if det["key"] == key:
                source_path = det["key"]
                file_url = det["data"]
                source_path = ".".join([source_path, "jpg"])
                # source_path = os.path.join(path_schema.PATH_FIGURE_DIR, source_path)
                fig_paths.append([source_path, file_url])

                # print(source_path, file_url)

                return fig_paths
    print("[bold red] Unable to find the figure")

    return fig_paths


# def give_fig_urls(details):

#     fig_paths = None

#     if details is not None:

#         details = details[0]["data"]
#         if "figure" in details.keys():
#             fig_paths = []

#             fig_details_all = details["figure"]

#             for fig_key, path in fig_details_all.items():
#                 source_path = fig_details_all[fig_key]["path"]["source_path"]
#                 source_url = fig_details_all[fig_key]["path"]["source_type"][
#                     "public_url"
#                 ]
#                 file_url = urljoin(source_url, source_path)
#                 fig_paths.append([source_path, file_url])

#     return fig_paths


# def delete(name:str, model_name:str,  model_version:str='latest') -> str:
#     '''`delete()` deletes an figure from a model

#     Parameters
#     ----------
#     name : str
#         The name of the figure you want to delete.
#     model_name : str
#         The name of the model you want to delete the figure from
#     model_version: str
#         The version of the model

#     '''

#     user_token = get_token()
#     org_id = get_org_id()

#     url_path_1 = '{}/project/{}/model/{}/{}/figure/{}/delete'.format(org_id, project_id, model_name, model_version, name)
#     url = urljoin(BASE_URL, url_path_1)


#     headers = {
#         'Content-Type': 'application/x-www-form-urlencoded',
#         'Authorization': 'Bearer {}'.format(user_token)
#     }


#     # figure_details = details(model_name=model_name, figure=figure)

#     # if figure_details is None:
#     #     print('[bold red] Unable to find figure details')
#     #     return


#     response = requests.delete(url, headers=headers)


#     if response.status_code == 200:
#         print(f"[bold green]figure has been deleted")

#     else:
#         print(f"[bold red]Unable to delete figure")

#     return response.text




from sklearn.metrics import accuracy_score


class Accuracy():

    def __init__(self):
        self.name = 'accuracy'
        self.input_type = 'int'
        self.output_type = None
        self.kwargs = None
        

    def parse_data(self, data):
        
        return data



    def compute(self, references, predictions, normalize=True, sample_weight=None, **kwargs):

        score = accuracy_score(y_true=references, y_pred=predictions, normalize=normalize,
                                sample_weight=sample_weight)
        
        score = {
            self.name : float(score)
            }
 

        return score